From mathcomp Require Import all_ssreflect all_algebra fingroup perm zmodp.

Require Import Psatz.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(* transforms a goal that is a boolean conjunction of comparisons into
  a collection of goals that are all coq_nat comparisons. *)
Ltac to_lia := repeat (apply/andP; split); apply/leP.

(* apply lia on a goal, after converting all arithmetic operations to
  operations that lia knows about. *)
Ltac ssr_lia := rewrite -?(plusE, minusE, multE, subn1); lia.

Lemma mem_iter_orbit (T : finType) (f : T -> T) p i :
  iter i f p \in orbit f p.
Proof. by rewrite -fconnect_orbit fconnect_iter. Qed.

Lemma iter_lt_order_neq (T : finType) f k l (x : T) :
  (k < l < fingraph.order f x)%N -> iter k f x != iter l f x.
Proof.
move => /andP [kl lo]; apply/eqP=> abs.
have ko : (k < fingraph.order f x)%N by apply: ltn_trans lo.
by move: (kl); rewrite ltn_neqAle -(findex_iter lo) -abs findex_iter // eqxx.
Qed.

Lemma subset_orbit (T : finType) (f : T -> T) (x y : T) :
   y \in orbit f x -> {subset (orbit f y) <= (orbit f x)}.
Proof.
move=> yo z; rewrite -!fconnect_orbit; apply: connect_trans.
by rewrite fconnect_orbit.
Qed.

Record cc_fun (T : Type) := {
  ccw : T -> T -> T -> bool;
}.

Record cc_system (T : eqType) (f : cc_fun T) := {
  Knuth_1 : forall a b c, ccw f a b c -> ccw f b c a;
  Knuth_2 : forall a b c, ccw f a b c -> ~~ ccw f b a c;
  Knuth_3 : forall a b c, uniq [:: a; b; c] -> ccw f a b c \/ ccw f b a c;
  Knuth_4 : forall a b c d, ccw f a b d -> ccw f b c d -> ccw f c a d ->
      ccw f a b c;
  Knuth_5 : forall a b c d e, ccw f a b c -> ccw f a b d -> ccw f a b e ->
    ccw f a c d -> ccw f a d e -> ccw f a c e;
  Knuth_5' : forall a b c d e, ccw f a b c -> ccw f a b d -> ccw f a b e ->
    ccw f b c d -> ccw f b d e -> ccw f b c e
}.

Record map_funs (T : Type) := {
  sort : Type;
  map0 : sort;
  eval : sort -> T -> T;
  update : sort -> T -> T -> sort;
  map_size : sort -> nat
}.

Record map_system (T : eqType) (funs : map_funs T) := {
  eval_update1 : forall (m : sort funs) x y, eval (update m x y) x = y;
  eval_update_dif : forall (m : sort funs) x x' y, x != x' ->
    eval (update m x y) x' = eval m x';
  map_sizeP : forall (m : sort funs) (x : T)(p : seq T),
    fpath (eval m) x p -> uniq (x :: p) ->
            size (x :: p) <= map_size m
}.

Lemma nth_orbit (T : finType) (d : T) (f : T -> T) (e : T) (i : nat):
  i < size (orbit f e) -> nth d (orbit f e) i = iter i f e.
Proof.
rewrite size_orbit /orbit => ilt. 
by rewrite (set_nth_default e) ?size_traject ?nth_traject.
Qed.

Section seq_algorithm.

Variables (P : Type) (p0 : P) (ccf : cc_fun P)
  (mf : map_funs P).

Notation ccw := (ccw ccf).
Notation update := update.
Notation map0 := (map0 mf).

Definition seq3 (s : seq P) :=
   let a := nth p0 s 0 in
   let b := nth a s 1 in
   let c := nth b s 2 in
   if ccw a b c then [:: a; b; c] else [:: b; a; c].


Definition map_cycle3 (s : seq P) :=
  match s with
  | [:: a; b; c] => update (update (update map0 c a) b c) a b
  | _ => map0
  end.

Definition inside_triangle_seq (trs : seq P) (p : P) :=
  match trs with
  | [:: a; b; c] => ccw a b p && ccw b c p && ccw c a p
  | _ => false
  end.

Fixpoint pick_triangle_seq (trs : seq (seq P)) (p : P) :=
  match trs with
  | tr :: trs =>
    if inside_triangle_seq tr p then Some(tr, trs)
    else
       match pick_triangle_seq trs p with
       | Some (v, others) => Some (v, tr::others)
       | None => None
       end
  | _ => None
  end.

Definition inside_triangles tr (p : P):=
  match tr with
  | [:: a; b; c] => [:: [:: p; a; b]; [:: p; b; c]; [:: p; c; a]]
  | _ => [::]
  end.

Fixpoint find_next_purple_map (p : P) (map : sort mf) side n b :=
  match n with
    0 => b
  | S k => if (ccw b (eval map b) p) == side then
             b
           else find_next_purple_map p map side k (eval map b)
  end.

Definition find_purple_points_map p (m : sort mf) (e : P) : P * P :=
  let n := map_size m in
  if ccw e (eval m e) p then
   let pp1 := find_next_purple_map p m false n (eval m e) in
     (pp1, find_next_purple_map p m true n (eval m pp1))
  else
   let pp2 := find_next_purple_map p m true n (eval m e) in
     (find_next_purple_map p m false n (eval m pp2), pp2).

Fixpoint add_red_triangles p (l : seq P) tr :=
  match l with
  | a :: (b :: _) as l' => 
    if ccw b a p then
      add_red_triangles p l' ([:: b; a; p] :: tr)
    else
      tr
  | _ => tr
  end.

Fixpoint map_orbit
  (eq_f : P -> P -> bool)
  (m : sort mf) (e stopper : P) (fuel : nat) trailer :
  seq P :=
  match fuel with
  | O => trailer
  | S k =>
    if eq_f (eval m e) stopper then
      e::trailer
    else
       e::map_orbit eq_f m (eval m e) stopper k trailer
  end.

Definition new_map (m : sort mf) (a b p : P) :=
  update (update m p b) a p.

Fixpoint naive5 eq_f (s : seq P) :
   seq (seq P) * (sort mf) * P :=
   match  s with
    nil =>  (nil, map0, p0)
  | p :: tl => 
    if (size s <= 3)%nat then
      let tr1 := seq3 s in
      ([:: tr1], map_cycle3 tr1, head p0 tr1)
    else 
      let '(tr, chf, bp) := naive5 eq_f tl in
      if pick_triangle_seq tr p is Some (t, trmt) then
       (inside_triangles t p ++ trmt, chf, bp)
      else
        let (a, b) := find_purple_points_map p chf bp in
          (add_red_triangles p
            (map_orbit eq_f chf a a (map_size chf) [:: a]) tr,
           new_map chf a b p, p)
  end.

End seq_algorithm.

Section proofs_in_finType.

Variables (P : finType) (p0 : P) (ccf : cc_fun P) (ccs : cc_system ccf)
  (mf : map_funs P) (ms: map_system mf).


Notation ccw := (ccw ccf).
Notation inside_triangle_seq := (@inside_triangle_seq P ccf).
Notation seq3 := (@seq3 P p0 ccf).
Notation map_cycle3 := (@map_cycle3 P mf).
Notation naive5 := (@naive5 P p0 ccf mf eq_op).
Notation pick_triangle_seq := (@pick_triangle_seq P ccf).
Notation inside_triangles := (@inside_triangles P).
Notation Knuth_1 := (Knuth_1 ccs).
Notation Knuth_2 := (Knuth_2 ccs).
Notation Knuth_3 := (Knuth_3 ccs).
Notation Knuth_4 := (Knuth_4 ccs).
Notation Knuth_5 := (Knuth_5 ccs).
Notation Knuth_5' := (Knuth_5' ccs).


Definition subE := (subUset, sub1set, inE, eqxx, orbT, andbT).

Definition ccw_seq (s : seq P) :=
  match s with [:: a; b; c] => ccw a b c | _ => false end.

Definition inh f p x :=
  forall e, e \in orbit f p -> ccw e (f e) x.

Definition inh' boundary x :=
  path.cycle (fun u v => ccw u v x) boundary.

Lemma inh_inh' f p x :
  fcycle f (orbit f p) ->
  reflect (inh f p x) (inh' (orbit f p)  x).
Proof.
move=> cyco.
apply/introP.
  rewrite /inh'; move=> h' y yin.
  have := next_cycle cyco yin => /= /eqP fyq.
  by have := next_cycle h' yin; rewrite <- fyq.
move/negP=> h' h; apply/h'=> {h'}; rewrite /inh' /path.cycle.
move: h cyco; rewrite /inh /orbit; rewrite -orderSpred /= => h cyco.
apply/(pathP p)=> i szi.
move/pathP: cyco => /(_ p _ szi) /= /eqP <-; apply:h.
rewrite -rcons_cons nth_rcons; case: ifP => extra; first by apply: mem_nth.
by case: ifP=> _; rewrite inE eqxx.
Qed.

Definition triangles_prop (tr : seq (seq P)) :=
  forall t, t \in tr -> size t = 3 /\ uniq t /\ ccw_seq t.

Definition hull_prop (inp : seq P) (m : sort mf) (p : P) :=
  forall q, q \in orbit (eval m) p ->
     forall x, x \in inp -> x != q -> x != eval m q ->
     ccw q (eval m q) x.

Definition hull_prop' (inp : seq P) (boundary : seq P) :=
  0 < size boundary /\
  path.cycle
    (fun x y =>
       [forall z, (z \in inp) ==> (z != x) ==> (z != y) ==> ccw x y z])
    boundary.

Lemma hull_prop_equiv (inp : seq P) (m : sort mf) (p : P) :
  fcycle (eval m) (orbit (eval m) p) ->
  hull_prop inp m p <-> hull_prop' inp (orbit (eval m) p).
Proof.
move=> cyco; split.
  move=> hp; rewrite /hull_prop' /orbit /path.cycle; rewrite -orderSpred /=.
  split; first by [].
  apply/(pathP p) => i szi; apply/forallP=> x; apply/implyP=> xin.
  apply/implyP=> xni; apply/implyP=> xni1.
  move:(cyco); rewrite /orbit/path.cycle; rewrite -orderSpred /= => /(pathP p).
  move=> /(_ i szi) /= /eqP eqi; rewrite -eqi; apply: hp; rewrite ?eqi //.
  rewrite /orbit -orderSpred /=; rewrite -rcons_cons nth_rcons.
  case: ifP=> // cnd; first by apply: mem_nth.
  by case: ifP; rewrite inE eqxx.
move=> [sn0 hp'] x xin y yin ynx ynfx.
move: hp' cyco xin; rewrite /path.cycle /orbit; rewrite -orderSpred /=.
set pp' := traject _ _ _.
move=> /pathP hp' /pathP cyco xin.
set i := index x (p :: pp').
have szi : (i < size (p :: pp')) by rewrite index_mem.
have szi' : (i < size (rcons pp' p)) by rewrite size_rcons.
move: (cyco p i szi'); rewrite /= => /eqP evq.
move: evq; rewrite -rcons_cons nth_rcons => evq.
move/(nth_index p) : (xin); rewrite -/i => vx.
move: (hp' p i szi') => /forallP /(_ y); rewrite yin -rcons_cons.
by rewrite -evq nth_rcons szi vx ynx ynfx !implyTb.
Qed.

Definition correct_triangulation (inp : seq P) (tr : seq (seq P))
   (m : sort mf) (p : P) :=
  triangles_prop tr /\
  (flatten tr =i inp) /\
  (orbit (eval m) p \subset inp) /\
  hull_prop inp m p /\
  (forall t1 t2 p, t1 \in tr -> t2 \in tr ->
    inside_triangle_seq t1 p ->
    inside_triangle_seq t2 p -> t1 = t2) /\
  (forall q, inh (eval m) p q ->
     (q \in inp \/ exists t, t \in tr /\ inside_triangle_seq t q)) /\
  uniq tr /\
  (fcycle (eval m) (orbit (eval m) p)).

Definition correct_triangulation' (inp : seq P) (tr : seq (seq P))
   (boundary : seq P) :=
  triangles_prop tr /\
  (flatten tr =i inp) /\
  (boundary \subset inp) /\
  hull_prop' inp boundary /\
  (forall t1 t2 p, t1 \in tr -> t2 \in tr -> 
    inside_triangle_seq t1 p ->
    inside_triangle_seq t2 p -> t1 = t2) /\
  (forall q, path.cycle (fun x y => ccw x y q) boundary ->
     (q \in inp \/ exists t, t \in tr /\ inside_triangle_seq t q)) /\
  uniq tr /\
  0 < size boundary /\
  uniq boundary.

Lemma correct_triangulation'_complete inp tr m p :
  correct_triangulation inp tr m p ->
  correct_triangulation' inp tr (orbit (eval m) p).
Proof.
move=> [] trp [] allp [] oin [] hp [] ut [] et [] utr cyco.
split;[exact trp | split; [exact allp | split; [exact oin |]]].
split;[rewrite -hull_prop_equiv  //| split; [exact ut | ]].
split;[move=> q cnq; apply et; apply/inh_inh' => // | split;[exact utr |]].
by split;[rewrite size_orbit -orderSpred | rewrite orbit_uniq].
Qed.

Definition seq_to_map (T : eqType) s (mf : map_funs T) :=
  foldr (fun (x : T) m => update m x (next s x)) (map0 mf) s.

Lemma eval_seq_to_map_cat (T : eqType) mf' (s1 s2 s : seq T) x :
  map_system mf' -> x \notin s1 ->
  eval (foldr (fun (x : T) m => update m x (next s x)) (map0 mf') (s1++s2)) x =
  eval (foldr (fun (x : T) m => update m x (next s x)) (map0 mf') s2) x.
Proof.
move=> ms'; move: x; elim: s1 => [// | a s1 IH].
move=> x xnas; rewrite /= eval_update_dif //; last first.
  by move: xnas; rewrite !inE negb_or eq_sym => /andP[].
by apply: IH; move: xnas; rewrite !inE negb_or => /andP[].
Qed.

Lemma eval_seq_to_map s x :
  uniq s -> x \in s -> eval (seq_to_map s mf) x = next s x.
Proof.
move=> uns xin.
rewrite -(nth_index x xin); set j := index _ _.
have szj : j < size s by rewrite /j index_mem.
move: (uns); rewrite /seq_to_map.
rewrite -[X in uniq X -> eval (foldr _ _ X) _ = _](cat_take_drop j).
rewrite (drop_nth x szj)=> uns'; rewrite eval_seq_to_map_cat //; last first.
  move: uns'; rewrite cat_uniq => /andP[] unt /andP[] disj und.
  by apply/negP=> A; case/negP: disj=> /=; rewrite A.
by rewrite /= eval_update1.
Qed.

Lemma order_seq_to_map a s :
  uniq (a :: s) -> fingraph.order (eval (seq_to_map (a :: s) mf)) a =
  size (a :: s).
Proof.
move=> uas.
rewrite /fingraph.order.
move: (uas) => /card_uniqP <-; apply: eq_card=> x; apply/idP/idP.
  rewrite inE fconnect_orbit=> /trajectP [i _]; move: i x.
  elim=> [ /= | i IH] x; first by move=> xi; rewrite xi !inE eqxx.
  rewrite iterS; set y := iter _ _ _ => xq; have IH' := IH y erefl.
  move/(nth_index a)/esym: (IH')=> yq.
  have iys : (index y (a :: s)) < size (a :: s) by rewrite index_mem.
  by rewrite xq eval_seq_to_map // mem_next.
move=> xin; rewrite -(nth_index a xin); move: (xin); rewrite -index_mem.
elim: (index x (a :: s)) => [ //= | i IH]; first by rewrite inE connect0.
move=> szi; have szim : i < size (a:: s) by rewrite (ltn_trans (ltnSn _)).
have /IH IH' := szim.
have := next_nth (a :: s) (nth a (a :: s) i).
rewrite mem_nth // nthK // -[nth a s i]/(nth a (a :: s) i.+1) => nextq.
rewrite -nextq -eval_seq_to_map // ?mem_nth //.
by apply (connect_trans IH'); rewrite fconnect1.
Qed.

Lemma orbit_seq_to_map (s : seq P) :
  0 < size s ->
  uniq s -> orbit (eval (seq_to_map s mf)) (head p0 s) = s.
Proof.
case: s => [// | a s] _.
move=> uas; rewrite /orbit [head _ _]/= order_seq_to_map //.
rewrite -[RHS](take_size (a :: s)).
apply: (@eq_from_nth _ a).
  by rewrite size_traject size_take; case: ifP.
rewrite size_traject=> i szi; rewrite nth_traject // nth_take //.
elim: i szi => [// | i IH] szi.
have szi' : i < size (a :: s) by apply: ltnW.
by rewrite iterS IH // eval_seq_to_map // ?next_nth // ?mem_nth// nthK.
Qed.

Lemma correct_triangulation_sound inp tr hull :
  correct_triangulation' inp tr hull ->
  correct_triangulation inp tr (seq_to_map hull mf) (head p0 hull).
Proof.
move=> [] trp [] allp [] oin [] hp [] ut [] et [] utr [] szh hun.
have cyco : fcycle (eval (seq_to_map hull mf))
    (orbit (eval (seq_to_map hull mf)) (head p0 hull)).
  rewrite orbit_seq_to_map // cycle_from_next //.
  by move=> x xin /=; apply/eqP; apply: eval_seq_to_map.
split;[exact: trp | split; [exact: allp | ]].
split; first by rewrite orbit_seq_to_map.
split; first by rewrite hull_prop_equiv // ?orbit_seq_to_map.
split;[exact: ut | split].
  move=> q /inh_inh'; rewrite orbit_seq_to_map //.
  rewrite cycle_from_next //; last by move=> x xin /=; rewrite eval_seq_to_map.
  by move=> /(_ isT) qh; apply: et.
by split.
Qed.

Lemma inh_orbit f p q : fcycle f (orbit f p) -> q \in orbit f p ->
  forall x, inh f p x <-> inh f q x.
Proof.
move=> cyc qin x; have [k oq] := orbit_rot_cycle cyc (orbit_uniq _ _) qin.
by rewrite /inh; split=> inh e; move: (inh e); rewrite oq mem_rot.
Qed.

Lemma correct_triangulation_perm i i' tr m p p' :
  perm_eq i i' -> p' \in (orbit (eval m) p) ->
  correct_triangulation i tr m p ->
  correct_triangulation i' tr m p'.
Proof.
move=> ii' pp' [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
split; first by apply: trP.
split.
  by move=> z; rewrite ctr; apply: perm_mem.
have [k oq] := orbit_rot_cycle cyc (orbit_uniq _ _) pp'.
have oqi : orbit (eval m) p' =i orbit (eval m) p.
  by move=> z; rewrite oq mem_rot.
split.
  apply/subsetP=> z; rewrite oqi; move/subsetP: incP=> A /A.
  by rewrite (perm_mem ii').
split.
  move=> q qin x xin'; have xin : x \in i by rewrite (perm_mem ii').
  by apply: (hulP q _ x xin); rewrite -oqi.
split=> //.
split.
  move=> q; rewrite -(inh_orbit cyc pp') => qin.
  case : (cov _ qin) => [qinp | ex_t].
    by left; rewrite -(perm_mem ii').
  right; exact ex_t.
by rewrite oq rot_cycle.
Qed.

Lemma seq3_perm s : size s = 3 -> perm_eq s (seq3 s).
Proof.
case: s => [ | a [| b [| c []]]]// _.
rewrite /seq3 ![nth _ _ _]/=; case ccw => //.
by rewrite (perm_catCA [::a] [::b] [::c]).
Qed.

Lemma seq3_ccw s : size s = 3 -> uniq s -> ccw_seq (seq3 s).
Proof.
case s=> [| a [| b [| c []]]] // _ unabc.
rewrite /seq3 ![nth _ _ _]/=; case test: (ccw a b c) => //.
by move: (Knuth_3 unabc)=> /orP; rewrite test orFb.
Qed.

Lemma orbit0 f (x : P) : x \in orbit f x.
Proof.  by rewrite -fconnect_orbit connect0. Qed.

Notation eval_update_dif := (@eval_update_dif P mf ms).
Notation eval_update1 := (@eval_update1 P mf ms).
Notation map_sizeP := (@map_sizeP P mf ms).
Notation find_purple_points_map := (@find_purple_points_map P ccf mf).

Lemma map_cycle3_value a b c : uniq [:: a; b; c] ->
  eval (map_cycle3 [:: a; b; c]) a = b /\
  eval (map_cycle3 [:: a; b; c]) b = c /\
  eval (map_cycle3 [:: a; b; c]) c = a.
Proof.
move=> unabc; rewrite /map_cycle3 eval_update1; split; first by [].
have /andP [/andP [anb anc ] bnc] : (a != b) && (a != c) && (b != c).
  by move: unabc; rewrite /= !inE negb_or andbT.
rewrite eval_update_dif // eval_update1 eval_update_dif // eval_update_dif //.
by rewrite eval_update1.
Qed.

Lemma orbit_map_cycle3 a b c: uniq [:: a; b; c] ->
  orbit (eval (map_cycle3 [:: a; b; c])) a = [:: a; b; c].
Proof.
move=> unabc; move: (map_cycle3_value unabc).
set m := map_cycle3 _ => [][va [vb vc]].
have /andP [/andP [anb anc] bnc] : (a != b) && (a != c) && (b != c).
  by move: unabc; rewrite /= !inE negb_or andbA andbT.
have size : fingraph.order (eval m) a = 3.
  rewrite /fingraph.order -[3]/(size [:: a; b; c]) -(card_uniqP (unabc)).
  apply: eq_card => z; apply/connectP/idP.
    case=> x; have : a \in [:: a; b; c] by rewrite !inE eqxx.
    elim: x {1 3 4}a => [| e' tl IH] e ein; first by move=> _ ->; exact ein.
    set C := (_ \in _); rewrite /= => /andP []; move: ein; rewrite /= !inE. 
    move=> /orP [ | /orP []] /eqP ->; rewrite ?va ?vb ?vc => vv;
    (have : e' \in [:: a; b; c] by rewrite -(eqP vv) !inE eqxx ?orbT);
    by rewrite /C; apply: IH.
  rewrite !inE => /orP [ | /orP [] ] /eqP ->; first by exists [::].
    by exists [:: b] => //=; rewrite va eqxx. 
  by exists [:: b; c] => //=; rewrite va vb ?eqxx.
by rewrite /orbit size /= va vb.
Qed.

Lemma Knuth_1_1 a b c : ccw a b c -> ccw c a b.
Proof. by move=> *; apply/Knuth_1/Knuth_1. Qed.

Lemma base_case inp tr m p :
  size inp = 3 -> uniq inp -> 
  naive5 inp = (tr, m, p) ->
  correct_triangulation inp tr m p.
Proof.
case: inp => [ | a [| b [| c [|]]]] // s3 unabc.
rewrite /naive5 /= => [][] <- <- <-; set l := seq3 _.
have ps3 : perm_eq [:: a; b; c] l by apply: seq3_perm.
have s3' : size l = 3 by rewrite -(perm_size ps3). 
have unl : uniq l by rewrite -(perm_uniq ps3).
have o0 := orbit0 (eval (map_cycle3 l)) (head p0 l).
move: (ps3); rewrite perm_sym => ps3'.
apply: (correct_triangulation_perm ps3' o0).
have := seq3_ccw s3 unabc; rewrite -/l.
move: (l) s3' unl=> {ps3' o0 ps3 l unabc s3 a b c}.
case=> [| a [| b [| c []]]] // s3 unabc abc.
split; first by move=> t; rewrite !inE => /eqP-> //.
split; first by [].
split; first by rewrite orbit_map_cycle3.
split.
  move=> q; rewrite orbit_map_cycle3 //.
  by have [va [vb vc]] := map_cycle3_value unabc; rewrite !inE;
  move => /orP [ | /orP []] /eqP ->; rewrite ?va ?vb ?vc => x;
  rewrite !inE => /orP [ | /orP []] /eqP ->;
  rewrite ?eqxx // => _ _; rewrite Knuth_1 // Knuth_1 //.
split; first by move=> t1 t2 p'; rewrite !inE => /eqP -> /eqP ->.
rewrite /inh orbit_map_cycle3 //= -?[update _ _ _]/(map_cycle3 [:: a; b; c]).
split.
  move=> q qin; right; exists [:: a; b; c]; rewrite !subE; split; first by [].
  move: (map_cycle3_value unabc) => [va [vb vc]].
  rewrite /inside_triangle_seq; move: (qin a) (qin b) (qin c).
  by rewrite !subE va vb vc => /(_ isT) -> /(_ isT) -> /(_ isT) ->.
by move: (map_cycle3_value unabc) => [-> [-> ->]]; rewrite !eqxx.
Qed.

Lemma pick_triangle_seq_someP (tr tr' : seq (seq P)) t (q : P) :
  pick_triangle_seq tr q = Some (t, tr') ->
  inside_triangle_seq t q && perm_eq (t :: tr') tr.
Proof.
elim: tr tr' t => [ | t1 trtl IH] tr' t //=.
have [ ins | nins] := boolP (inside_triangle_seq t1 q).
  by move=> [<- <-]; rewrite ins perm_refl.
case cmp : (pick_triangle_seq trtl q) => [[v1 vtr] |] //.
move=> [v1t t1vtrtr'].
have := IH _ _ cmp; rewrite -t1vtrtr' -v1t=> /andP[] -> peq.
by rewrite andTb (perm_catCA [:: v1] [:: t1] vtr) /= perm_cons.
Qed.

Lemma pick_triangle_seq_noneP (tr : seq (seq P)) (q : P) :
  pick_triangle_seq tr q = None ->
  forall t, t \in tr -> ~~inside_triangle_seq t q.
Proof.
elim: tr q => [ | t1 trtl IH] q //=.
have [ ins | nins] := boolP (inside_triangle_seq t1 q) => //.
case cmp : (pick_triangle_seq trtl q) => [[v1 vtr] |] // _.
by move=> t; rewrite !inE => /orP [/eqP -> // | tin]; apply: IH.
Qed.

Lemma mem_orbit1 (T : finType) (f : T -> T) (x y : T) :
  y \in orbit f x -> f y \in orbit f x.
Proof.
by rewrite -!fconnect_orbit => A; apply/(connect_trans A)/fconnect1.
Qed.

Lemma uniq_hull2 inp m p e:
  2 < size inp -> uniq inp ->
  e \in orbit (eval m) p ->
  hull_prop inp m p ->
  uniq [:: e; eval m e; eval m (eval m e)].
Proof.
move=> s3 uni ein hulP.
have [q [qin qdif]] : exists q, q \in inp /\ q \notin [:: e; eval m e].
  move: s3 uni; case h : inp => [| a [| b [| c tl]]] // _.
  have [ang | aok] := boolP(a \in [:: e; eval m e]); last first.
    by exists a; rewrite aok !subE.
  have [bng | bok] := boolP(b \in [:: e; eval m e]); last first.
    by exists b; rewrite bok !subE.
  move=> U; exists c; rewrite !subE negb_or; split=> //; move: ang bng U.
  by rewrite /= !inE => /orP[] /eqP -> /orP [] /eqP ->;
     rewrite ?eqxx ?orbT ?andbF ?negb_or !andbA -!(eq_sym c) // =>
     /andP [] /andP [] /andP [] /andP [] /andP [] /andP [] => *;
     apply/andP; split.
have /andP [qne qnme] : (q != e) && (q != eval m e).
  by move: qdif; rewrite !subE negb_or.
have := hulP _ ein _ qin qne qnme => efq; move/Knuth_2: (efq)=> nfeq.
have enf : e != eval m e.
  by apply/eqP=> abs; move: efq nfeq; rewrite -abs=> ->.
have [/eqP qff | qnff] := boolP (q == eval m (eval m e)).
  by rewrite /= !subE negb_or enf -qff -!(eq_sym q) qne qnme.
have fein := mem_orbit1 ein.
have := hulP _ fein _ qin qnme qnff => feffq; move/Knuth_2: (feffq)=> n'.
rewrite /= !subE negb_or enf /=; apply/andP; split; apply/eqP=> abs.
  by move: n'; rewrite -abs efq.
by move: feffq n'; rewrite -abs=> ->.
Qed.

Lemma no_triangle_red_edge (inp : seq P) (tr : seq (seq P)) m p q :
  (forall t, t \in tr -> ~~ inside_triangle_seq t q) ->
  2 < size inp -> uniq inp ->
  q \notin inp ->
  correct_triangulation inp tr m p ->
  exists2 e, e \in orbit (eval m) p & ccw (eval m e) e q.
Proof.
move=> no_t s3 uni qnotin cot.
move: (cot) => [] trP [] ctr [] incP [] hulP [] no_i [] cov cyc.
have [] := boolP ([exists e, (e \in orbit (eval m) p) && ccw (eval m e) e q]).
  by move=> /existsP [e /andP [ein ered]]; exists e; rewrite ?ein ?ered.
rewrite negb_exists=> /forallP allblue; have inhq : (inh (eval m) p q).
  rewrite /inh=> e ein; move: (allblue e); rewrite ein andTb=> /negbTE eblue.
  have qno : q \notin orbit (eval m) p.
    by apply/negP=> abs; move/negP: qnotin; apply; rewrite (subsetP incP).
  have qne : q != e by apply/eqP=> abs; move: qno; rewrite abs ein.
  have qnfe : q != eval m e.
    by apply/eqP=> abs; move: qno; rewrite abs mem_orbit1.
  have /Knuth_3 : uniq [:: e; eval m e; q].
    rewrite /= !subE negb_or -!(eq_sym q) qnfe qne.
    have:= (uniq_hull2 s3 uni ein hulP) => /=.
    by rewrite !inE negb_or => /andP[] /andP [] ->.
  by move/orP; rewrite eblue orbF.
have := cov _ inhq; case; first by rewrite (negbTE qnotin).
by move=> [] t [] ttr qint; move: (no_t t ttr); rewrite qint.
Qed.

Lemma perm_flatten_mem (T : finType) (s1 s2 : seq (seq T)) :
  perm_eq s1 s2 ->
  flatten s1 =i flatten s2.
Proof.
elim: {s1} (size s1) {-2} s1 s2 (erefl (size s1))=> [ | n IH] s1 s2.
  move=> /eqP; rewrite size_eq0 perm_sym=> /eqP -> /perm_size/eqP.
  by rewrite size_eq0=> /eqP ->.
case: s1 => [| e s1] //= /eqP; rewrite eqSS => /eqP sz1.
move=> peq; have ein : e \in s2 by rewrite -(perm_mem peq) !subE.
have sz2 : size s2 = n.+1 by rewrite -(perm_size peq) /= sz1.
have szr : size (rem e s2) = n by rewrite (size_rem ein) sz2.
have peq' : perm_eq s2 (e :: rem e s2) by apply: perm_to_rem.
have : perm_eq (e :: s1) (e :: rem e s2) by apply/(perm_trans peq).
rewrite perm_cons => peq''.
case q2 : s2 => [| e' s2']; first by move: sz2; rewrite q2.
have [ee' | ene'] := boolP (e == e').
  rewrite (eqP ee') /= => z; rewrite !mem_cat; case zin: (z \in e')=> //=.
  by apply: IH=> //; move: peq''; rewrite q2 /= eq_sym ee'.
move: (ein) => /perm_to_rem peq3.
have ein2 : e \in s2 by rewrite -(perm_mem peq) inE eqxx.
have ein2' : e \in s2' by move: ein2; rewrite q2 inE (negbTE ene').
move=> z; rewrite [LHS]/= mem_cat; case zin : (z \in e) => /=.
  apply/esym=> /=.
  rewrite mem_cat; case zin' : (z \in e') => //=.
  have peq4 : perm_eq s2' (e :: rem e s2') by apply perm_to_rem.
  have sz' : size s2' = n by apply/eqP; rewrite -eqSS -sz2 q2.
  by rewrite (IH _ (e :: rem e s2') sz') //=; rewrite mem_cat zin.
have ein1 : e' \in (e :: s1) by rewrite (perm_mem peq) q2 !inE eqxx.
have ein1': e' \in s1 by move: ein1; rewrite !inE eq_sym (negbTE ene').
move: (ein1') => /perm_to_rem peq4.
rewrite (IH _  _ sz1 peq4) /= !mem_cat; case : (z \in e') => //=.
have -> : (z \in flatten (rem e' s1)) = (z \in (flatten (e :: rem e' s1))).
  by rewrite /= mem_cat zin.
apply/esym/IH.
  by apply/eqP; rewrite -eqSS -sz2 q2.
rewrite -(perm_cons e') -q2 perm_sym (perm_catCA [:: e'] [:: e]) /=.
by apply: perm_trans peq; rewrite perm_cons perm_sym; apply: perm_to_rem.
Qed.

Section in_existing_triangle.

Variables (inp t : seq P) (tr trmt : seq (seq P)) (m : sort mf) (p q : P).
Hypotheses (s3 : 2 < size inp) (uni : uniq (q :: inp)).
Hypothesis cot : correct_triangulation inp tr m p.
Hypotheses (qin : inside_triangle_seq t q) (ptr : perm_eq tr (t :: trmt)).

Lemma tint : t \in tr.
Proof. by rewrite (perm_mem ptr) !subE. Qed.

Lemma t_dec : 
  triangles_prop tr ->
  exists a b c, t = [:: a; b; c] /\ uniq [:: a; b; c] /\ ccw a b c.
Proof.
move=> trP.
by move/trP: tint=> []; case: (t) => [| a [| b [| c []]]] //; exists a, b, c.
Qed.

Lemma inside_triangles_seq1P :
  forall u, u \in inside_triangles t q -> size u = 3 /\ uniq u /\ ccw_seq u.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov cyc u.
case : (t_dec trP) => a [b [c [tq [unabc abc]]]]; rewrite tq.
have : q \notin [:: a; b; c].
  apply/negP; rewrite -tq => abs; move: uni => /=; rewrite andbC=> /andP[] _.
  by move/negP; apply; rewrite -ctr (perm_flatten_mem ptr) /= mem_cat abs.
move: (unabc) => /=; rewrite !inE andbT !negb_or -!andbA.
move => /andP[] anb /andP[] anc bnc /andP[] qna /andP [] qnb qnc.
move: qin; rewrite /inside_triangle_seq tq=> /andP[] /andP[] 
   /Knuth_1 /Knuth_1 abq /Knuth_1 /Knuth_1 bcq /Knuth_1 /Knuth_1 caq.
move => /orP [ | /orP []] /eqP -> /=;
  rewrite !inE negb_or ?qna ?qnb ?qnc ?anb ?bnc ?abq ?bcq ?caq //.
by rewrite eq_sym anc.
Qed.

Lemma flatten_inside_triangles : flatten (inside_triangles t q) =i q::t.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov cyc.
have [a [b [c [tq [ut abc]]]]] := t_dec trP; rewrite tq /= => z; rewrite !subE.
rewrite !(orbCA _ (z == q)) !orbA !orbb -!orbA ?orbb; congr orb.
rewrite -1?(orbC (z == a)) !(orbCA _ (z == a)) !orbA !orbb -!orbA ?orbb. 
by congr orb; rewrite orbA orbb.
Qed.

Lemma oriented_no_dup a b : ~~ccw a a b.
Proof using ccs.
by apply/negP=> A; move/Knuth_2: (A) => /negP; apply.
Qed.

Lemma oriented_diff c a b : ccw a b c -> a != b.
Proof using ccs.
move=> abc; apply/negP=> /eqP A; move: abc; rewrite A.
apply/negP/oriented_no_dup.
Qed.

Lemma inside_triangle_edge1 a b c d e f :
  ccw a b c -> ccw a b d -> ccw a b e -> ccw c d f -> ccw c f e -> ccw f d e ->
  ccw a d c -> ccw a e c -> ccw a b f.
Proof.
move=> abc abd abe cdf cfe fde adc aec.
have [cna [ena enc]] : c != a /\ e != a /\ e != c.
  move: (abc) (abe) (cfe)=> /Knuth_1_1/oriented_diff ->.
  by move=> /Knuth_1_1/oriented_diff -> /Knuth_1_1/oriented_diff.
have [dna dne] : d != a /\ d != e.
  by move: abd fde=> /Knuth_1_1/oriented_diff-> /Knuth_1/oriented_diff.
have [anc [cne and]] : a != c /\ c != e /\ a != d.
  by repeat split; rewrite eq_sym.
have cde : ccw c d e.
  by apply: (@Knuth_4 _ _ _ f); do 3 (apply: Knuth_1 => //).
have [ecd cae] : ccw e c d /\ ccw c a e by split; do 3 (apply: Knuth_1 => //).
have caf : ccw c a f.
  by apply: (Knuth_5' _ ecd); do 3 (apply: Knuth_1 => //).
have/Knuth_3 [dae | ade] : uniq [:: d; a; e].
    by rewrite /= !inE negb_or dna dne eq_sym ena.
  apply: (Knuth_5' _ cae); do 3 (apply: Knuth_1 => //).
  by apply/Knuth_1_1/(Knuth_5 _ ecd); do 3 (apply: Knuth_1 => //).
have [cad dec] : ccw c a d /\ ccw d e c by split; do 3 (apply/Knuth_1 => //).
apply: (Knuth_5' _ cad); do 3 (apply/Knuth_1 => //).
apply/Knuth_1_1/(Knuth_5 _ dec); do 3 (apply/Knuth_1 => //).
Qed.

Lemma inside_triangle_left_of_edge a b c d e f :
  ccw a b c -> ccw a b d -> ccw a b e -> ccw c d f -> ccw c f e -> ccw f d e ->
  ccw a b f.
Proof.
move=> abc abd abe cdf cfe fde.
have /andP[ /andP[ad dc] ca] : (a != d) && (d !=c) && (c != a).
  move/oriented_diff: cdf abd abc; rewrite eq_sym !(eq_sym a) => ->.
  by move/Knuth_1_1/oriented_diff=> -> /Knuth_1_1/oriented_diff.
have /andP[ /andP[ac ce] ea] : (a != c) && (c != e) && (e != a).
  move: cfe abe=> /Knuth_1_1/oriented_diff; rewrite eq_sym=> ->. 
  by rewrite eq_sym ca=> /Knuth_1_1/oriented_diff.
have /Knuth_3 [adc | dac] : uniq [:: a; d; c].
    by rewrite /= !inE negb_or ad ac dc.
  have /Knuth_3 [ace | cae] : uniq [:: a; c; e].
      by rewrite /= !subE negb_or ce ac eq_sym ea.
    apply: (inside_triangle_edge1 abe abc abd); do 3 (apply/Knuth_1 => //).
    by apply: (Knuth_5 abd abc abe).
  by apply: (inside_triangle_edge1 abc abd abe); do 3 (apply/Knuth_1 => //).
have ed :e != d.
  by move: fde=> /Knuth_1/oriented_diff; rewrite eq_sym.
have /Knuth_3 [aed | cae] : uniq [:: a; e; d].
    by rewrite /= !subE negb_or ed ad eq_sym ea.
  by apply: (inside_triangle_edge1 abd abe abc); do 3 (apply/Knuth_1=> //).
apply: (inside_triangle_edge1 abe abc abd); do 3 (apply/Knuth_1=> //).
by apply: (Knuth_5 abc abd); do 3 (apply/Knuth_1=> //).
Qed.

Lemma inside_triangle_in_hull z x :
  uniq inp ->
  triangles_prop tr ->
  flatten tr =i inp ->
  hull_prop inp m p ->
  z \in orbit (eval m) p ->
  inside_triangle_seq t x ->
  x != z -> x != eval m z -> ccw z (eval m z) x.
Proof.
move=> uni' trP ctr hulP zin xin xnz xnfz.
have znfz : z != eval m z.
  have := uniq_hull2 s3 uni' zin hulP.
  by rewrite /= !inE negb_or=> /andP[] /andP[].
have [a [b [c [tq [ut abc]]]]] := t_dec trP.
have tsub : forall u, u \in t -> u \in inp.
  by move=> u uint; rewrite -ctr (perm_flatten_mem ptr) /= mem_cat uint.
have /andP[/andP[ain bin] cin] : (a \in inp) && (b \in inp) && (c \in inp).
   by rewrite !tsub // tq !subE.
have[/andP[znotin fznotin] | ]:= boolP ((z \notin t) && (eval m z \notin t)).
  have /andP[/andP[zfa zfb] zfc] : 
    ccw z (eval m z) a && ccw z (eval m z) b && ccw z (eval m z) c.
    by rewrite !hulP //; apply/negP=> A; move: znotin fznotin; 
       rewrite -(eqP A) tq !subE.
  by apply: (inside_triangle_left_of_edge zfa zfb zfc); move: xin;
  rewrite tq /= => /andP[] /andP[] // => *; do 2 (apply/Knuth_1 => //).
have [/andP [zint fzint] _ | onlyone] := boolP ((z \in t) && (eval m z \in t)).
have gen : forall u, u \in rem z (rem (eval m z) t) -> ccw z (eval m z) u.
    move=> u uin; have uin1 : u \in rem (eval m z) t by apply/(mem_rem uin).
    have uin2 : u \in inp by apply/tsub/(mem_rem uin1).
    have /(mem_rem_uniq (eval m z)) nn : uniq t by rewrite tq.
    have /(mem_rem_uniq z) nn' : uniq (rem (eval m z) t).
      by apply: rem_uniq; rewrite tq.
    move: uin uin1; rewrite nn nn' !inE => /andP[] unz _ /andP[] unfz _.
    apply: (hulP z zin u uin2 unz unfz).
  have : ((z == a) && (eval m z == b)) || ((z == b) && (eval m z == c)) ||
         ((z == c) && (eval m z == a)).
    by move: zint fzint znfz ut gen; rewrite tq /= !inE negb_or
      => /orP[|/orP[]]/eqP -> /orP[|/orP[]]/eqP ->;
      rewrite !eqxx ?andbT ?orbT => //= _ /andP[] /andP[] /negbTE c1 /negbTE c2
        /negbTE c3; do 2 (rewrite /= ?c1 ?c2 ?c3); rewrite /= ?eqxx ?inE;
        set u' := (X in [:: X]); move=> /(_ u'); rewrite !inE eqxx /u';
        move=> /(_ isT);
        do 3 ((move=>/Knuth_2; rewrite abc //) || (move=> /Knuth_1)).
  by move=> /orP[ /orP [] |] /andP[] /eqP -> /eqP ->; move: xin;
     rewrite /inside_triangle_seq tq => /andP[] /andP[].
rewrite negb_and !negbK => /orP[zint | fzint].
  move: onlyone; rewrite zint /= => fznint.
  have [b' [c' /andP[] /andP[] /andP[] /andP[] zbx czx zfzb zfzc czb]] : 
   exists b' c', ccw z b' x && ccw c' z x && ccw z (eval m z) b' &&
   ccw z (eval m z) c' && ccw c' z b'.
    by move: zint xin abc; rewrite tq /= !inE => /orP [| /orP[]] /eqP <-;
    set b1 := (X in ccw z X _); set c1 := (X in ccw X z _) => /andP[]/andP[];
    move=> A B C; exists b1, c1; rewrite ?A ?B ?C /= !hulP //;
    move/oriented_diff: A; move/oriented_diff: B; move/oriented_diff: C;
    rewrite !(eq_sym _ z) /c1 /b1=> // _ _ _; try(apply/eqP=> A; move: fznint;
    rewrite -A tq !subE //); do 3 (apply/Knuth_1 => //).
  move: zfzc=> /Knuth_1_1=> czfz.
  by apply: (Knuth_5' czfz czb czx zfzb zbx).
move: onlyone; rewrite fzint andbT => znint.
have [b' [c' /andP[] /andP[] /andP[] /andP[] fzbx cfzx zfzb zfzc cfzb]] : 
 exists b' c', ccw (eval m z) b' x && ccw c' (eval m z) x && 
    ccw z (eval m z) b' &&
    ccw z (eval m z) c' && ccw c' (eval m z) b'.
  by move: fzint xin abc; rewrite tq /= !inE => /orP [| /orP[]] /eqP <-;
  set b1 := (X in ccw (eval m z) X _); set c1 := (X in ccw X (eval m z) _);
  move=> /andP[]/andP[] A B C; exists b1, c1; rewrite ?A ?B ?C /= !hulP //;
  move/oriented_diff: A; move/oriented_diff: B; move/oriented_diff: C;
  rewrite !(eq_sym _ (eval m z)) /c1 /b1=> // _ _ _;
  try(apply/eqP=> A; move: znint;
  rewrite -A tq !subE //); do 3 (apply/Knuth_1 => //).
apply/Knuth_1_1; move: cfzb zfzb=> /Knuth_1 fzbc /Knuth_1 fzbz.
move: cfzx zfzc=> /Knuth_1 fzxc /Knuth_1 fzcz.
by apply: (Knuth_5 fzbx fzbc fzbz fzxc fzcz).
Qed.

Lemma uniq_take (T : eqType) (n : nat) (s : seq T) :
    uniq s -> uniq (take n s).
Proof.
by rewrite -[X in uniq X](cat_take_drop n s); rewrite cat_uniq => /andP[].
Qed.

Lemma inside_sub_triangle a b c d e :
  inside_triangle_seq [:: a; b; c] d ->
  inside_triangle_seq [:: a; b; d] e ->
  inside_triangle_seq [:: a; b; c] e.
Proof.
move=> /= /andP[] /andP[] abd bcd cad /andP[] /andP[] abe bde dae.
rewrite abe.
have abc : ccw a b c by apply/(Knuth_4 abd bcd cad).
have -> : ccw b c e by apply/(Knuth_5' abc abd abe bcd bde).
have -> // : ccw c a e.
by apply/Knuth_1_1/(Knuth_5 abe abd abc); apply/Knuth_1.
Qed.

Lemma inside_triangle_seq_K1 a b c d :
  inside_triangle_seq [:: a; b; c] d = inside_triangle_seq [:: b; c; a] d.
Proof. by rewrite /= -andbA andbC. Qed.

Lemma sub_triangle_unique e (j : 'I_3) :
  triangles_prop tr ->
  inside_triangle_seq (q :: (take 2 (rot j t))) e ->
  ~~inside_triangle_seq (q :: (take 2 (rot (j + 1)%R t))) e.
Proof.
move=> trP; have [a [b [c [tq [unabc abc]]]]] := t_dec trP.
by rewrite tq; case: j => [[| [| [| ?]]] ?] //= =>
  /andP[] /andP[] _ _ /Knuth_2/negbTE ->.
Qed.

Lemma inside_triangles_index u :
  triangles_prop tr ->
  (u \in inside_triangles t q) =
  [exists j : 'I_3, u == q :: (take 2 (rot j t))].
Proof.
move=> trP; have [a [b [c [tq [unabc abc]]]]] := t_dec trP.
rewrite tq /inside_triangles; apply/idP/existsP.
  rewrite !inE=> /orP [| /orP[]] /eqP ->;
  (exists 0%R; rewrite /= eqxx //) || (exists 1%R; rewrite /= eqxx //) ||
  (exists ((-1)%R); rewrite /= eqxx //) || idtac.
by move=> [ [ [| [| [| ?]]] ?] // /eqP ->] /=; rewrite !subE.
Qed.

Lemma jdif_add1 (i j : 'I_3) : i != j -> (i == j + 1)%R || (j == i + 1)%R.
Proof.
by case: i => [ [| [| [| ?]]] ?] //; case: j => [ [| [| [| ?]]] ?] //.
Qed.

Lemma inside_triangle_seq_uniq (a b c d : P) :
  inside_triangle_seq [:: a; b; c] d -> uniq [:: a; b; c; d].
Proof.
move=> /andP[] /andP[] abd bcd cad; move: (abd) (bcd) (cad).
move=> /oriented_diff anb /oriented_diff bnc /oriented_diff cna.
move: (abd) (bcd) (cad)=> /Knuth_1/oriented_diff bnd.
move=>/Knuth_1/oriented_diff cnd /Knuth_1/oriented_diff and.
by rewrite /= !subE !negb_or anb and bnc bnd cnd eq_sym cna.
Qed.

Definition auto_uniq := (inE, negb_or, (fun x y z=> esym (andbA x y z))).

Lemma choose_sub_triangle a b c d e :
  d != e ->
  inside_triangle_seq [:: a; b; c] d ->
  inside_triangle_seq [:: a; b; c] e ->
  exists t, t \in inside_triangles [:: a; b; c] d /\
     inside_triangle_seq t e.
Proof.
move=> dne abcd abce; move: (abcd) => /inside_triangle_seq_uniq und.
move: (abce) => /inside_triangle_seq_uniq une.
move: abcd abce=> /andP[]/andP[] abd bcd cad /andP[]/andP[] abe bce cae.
suff :
  [|| inside_triangle_seq [:: a; b; d] e, inside_triangle_seq [:: b; c; d] e |
      inside_triangle_seq [:: c; a; d] e].
  by move=> /orP[| /orP[]]; rewrite -inside_triangle_seq_K1; set u := (_ :: _);
  exists u=> /=; rewrite /u !subE; split.
rewrite /inside_triangle_seq.
have /Knuth_3 [bde | dbe] : uniq [:: b; d; e].
    rewrite /= !negb_or dne !andbT; move: und une; rewrite /= !auto_uniq.
    by repeat ((move=> /andP[] A; rewrite ?A {A}) || move=> _).
  have /Knuth_3 [cde | dce] : uniq [:: c; d; e].
      rewrite /= !negb_or dne !andbT; move: und une; rewrite /= !auto_uniq.
      by repeat ((move=> /andP[] A; rewrite ?A {A}) || move=> _).
    have /Knuth_3 [ade | dae] : uniq [:: a; d; e].
        rewrite /= !negb_or dne !andbT; move: und une; rewrite /= !auto_uniq.
        by repeat ((move=> /andP[] A; rewrite ?A {A}) || move=> _).
      by move/Knuth_1/Knuth_2: cad; rewrite (@Knuth_5 d e a b c) //;
        do 3 (apply/Knuth_1 => //).
    by rewrite abe bde dae ?orbT.
  have /Knuth_3 [ade | dae] : uniq [:: a; d; e].
      rewrite /= !negb_or dne !andbT; move: und une; rewrite /= !auto_uniq.
      by repeat ((move=> /andP[] A; rewrite ?A {A}) || move=> _).
    by rewrite cae ade dce ?orbT.
  by rewrite abe bde dae ?orbT.
have /Knuth_3 [cde | dce] : uniq [:: c; d; e].
    rewrite /= !negb_or dne !andbT; move: und une; rewrite /= !auto_uniq.
    by repeat ((move=> /andP[] A; rewrite ?A {A}) || move=> _).
  by rewrite bce cde dbe ?orbT.
have /Knuth_3 [ade | dae] : uniq [:: a; d; e].
    rewrite /= !negb_or dne !andbT; move: und une; rewrite /= !auto_uniq.
    by repeat ((move=> /andP[] A; rewrite ?A {A}) || move=> _).
  by rewrite cae ade dce ?orbT.
by move/Knuth_1/Knuth_2: cae; rewrite (@Knuth_5 e d a b c) //;
  do 3 (apply/Knuth_1 => //).
Qed.

Lemma uniq_inside_triangles a b c d :
  uniq [:: a; b; c] ->
  uniq (inside_triangles [:: a; b; c] d).
Proof.
move=> abc; rewrite /=; rewrite !subE negb_or.
by apply/andP; split;[apply/andP; split | ]; apply/eqP=> [][] A B; move: abc;
  rewrite ?A ?B /= !subE.
Qed.

Lemma correct_inside_triangle :
  correct_triangulation (q :: inp) (inside_triangles t q ++ trmt) m p.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
have uni' : uniq inp by move: uni; rewrite /= => /andP[].
have [a [b [c [tq [ut abc]]]]] := t_dec trP.
split.
  move=> u; rewrite mem_cat => /orP[] utrmt; last first.
    by apply: trP; rewrite (perm_mem ptr) inE utrmt orbT.
  by apply: (inside_triangles_seq1P utrmt).
split.
  rewrite flatten_cat=> z; rewrite mem_cat flatten_inside_triangles !inE -orbA.
  congr orb; rewrite -[LHS]mem_cat -[_ ++ _]/(flatten (t :: trmt)).
  by rewrite -(perm_flatten_mem ptr); apply: ctr.
split.
  by apply/subsetP=> z /(subsetP incP); rewrite inE orbC => ->.
split.
  move=> z zin x xin xnz xnfz; move: xin; rewrite !inE => /orP []; last first.
    by move=> xin; apply: (hulP z zin x xin xnz xnfz).
  move=> /eqP xq; move: xnz xnfz; rewrite xq=> qnz qnfz.
  by apply: inside_triangle_in_hull.
split.
  move=> t1 t2 p1; rewrite !mem_cat.
  have gen: forall u, u \in inside_triangles t q -> inside_triangle_seq u p1 ->
           inside_triangle_seq t p1.
  move: qin; rewrite tq => qin'; have th := inside_triangle_seq_K1.
  by move=> u /orP [|/orP[|/orP[]]] // /eqP ->; rewrite th => A;
    repeat (apply: inside_sub_triangle A => // || rewrite // th).
  have ttr : t \in tr by rewrite (perm_mem ptr) !subE.
  have trmtsub : {subset trmt <= tr}.
    by move=> z zin; rewrite (perm_mem ptr) inE zin orbT.
  move=> /orP[t1sub | t1o] /orP[t2sub |t2o] p1t1 p1t2.
        move: t1sub t2sub; rewrite !inside_triangles_index //.
        move=> /existsP [i /eqP t1v] /existsP [j /eqP t2v]; rewrite t1v t2v.
        have [/eqP -> // | /jdif_add1 ad1] := boolP (i == j).
        by move: ad1 t1v t2v p1t1 p1t2 => /orP [] /eqP ->;
        set k := (X in (X + 1)%R); rewrite -/k => -> -> A B;
        move: (@sub_triangle_unique p1 k trP); rewrite ?B ?A => /(_ isT).
      move: (no_i _ _ _ ttr (trmtsub _ t2o) (gen _ t1sub p1t1) p1t2)=> A.
      by move: unitr; rewrite (perm_uniq ptr) A /= t2o.
    move: (no_i _ _ _ ttr (trmtsub _ t1o) (gen _ t2sub p1t2) p1t1)=> A.
    by move: unitr; rewrite (perm_uniq ptr) A /= t1o.
  by apply: (no_i _ _ _ (trmtsub _ t1o) (trmtsub _ t2o) p1t1 p1t2).
split.
  move=> p1 p1in; case: (cov p1 p1in) => [p1inp | [t' [t'in p1int']]].
    by left; rewrite inE p1inp orbT.
  have [/eqP tt' | tnott'] := boolP (t == t'); last first.
    right; exists t'; rewrite mem_cat p1int'; split => //; move: t'in.
    by rewrite (perm_mem ptr) inE eq_sym (negbTE tnott') /= orbC => ->.
  have [/eqP -> | p1nq] := boolP( p1 == q); first by left; rewrite !subE.
  right; move: p1nq p1int' qin; rewrite eq_sym -tt' tq=> qnp1 p1int qin'.
  have [t1 [t1in p1int1]]:= (choose_sub_triangle qnp1 qin' p1int).
  by exists t1; rewrite mem_cat t1in.
split=> //.
rewrite cat_uniq andbA; apply/andP; split; last first.
  by move: unitr; rewrite (perm_uniq ptr) cons_uniq => /andP[].
apply/andP; split; last first.
  have qnotin : q \notin inp by move: uni; rewrite cons_uniq=>/andP[].
  apply/negP=> /hasP [t' t'trmt /= t'ins]; case/negP: qnotin.
  rewrite -ctr (perm_flatten_mem ptr) /= mem_cat; apply/orP; right.
  have qt' : q \in t'.
    move: t'ins; rewrite tq /= !inE.
    by move => /orP [|/orP[]] /eqP ->; rewrite !subE.
  by apply/flattenP; exists t'.
by rewrite tq; apply: uniq_inside_triangles.
Qed.

End in_existing_triangle.

Section outside_hull.

Variables (inp : seq P) (tr : seq (seq P)) (m : sort mf) (p q : P).
Hypotheses (s3 : 2 < size inp) (uni : uniq (q :: inp)).
Hypothesis cot : correct_triangulation inp tr m p.

Let f := eval m.

Variable red_point : P.
Hypothesis redin : red_point \in orbit f p.
Hypothesis redP : ccw (f red_point) red_point q.


Lemma convex_hull_path_inj : { in orbit f p &, injective f}.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
move=> x y xo yo ff; have[/eqP // | ynx] := boolP(y == x).
have [xinp yinp] : x \in inp /\ y \in inp by split; apply:(subsetP incP).
have uni' : uniq inp by move: uni; rewrite cons_uniq=> /andP[].
have := hulP x xo y yinp ynx; rewrite -/f ff => xfyy.
have := uniq_hull2 s3 uni' yo hulP.
rewrite /= -/f !subE negb_or -andbA=>/andP[].
move/xfyy/Knuth_1/Knuth_2=> A _; move: A.
have := hulP y yo x xinp; rewrite eq_sym ynx -/f => yfyx.
have := uniq_hull2 s3 uni' xo hulP.
rewrite /= -/f !subE negb_or -andbA=>/andP[].
by rewrite ff; move=>/(yfyx isT)=> ->.
Qed.

Lemma same_orbit : {in orbit f p, forall x, orbit f x =i orbit f p}.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
move=> x xin; have [j ->]:= (orbit_rot_cycle cyc (orbit_uniq _ _) xin).
by move=> z; rewrite mem_rot.
Qed.

Lemma qnotiter a n : a \in orbit f p -> q != iter n f a.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
move=> ain; apply/negP=> /eqP A; move: uni.
by rewrite cons_uniq A (subsetP incP) // -(same_orbit ain) mem_iter_orbit.
Qed.

Lemma general_position_orbit x : 
  hull_prop inp m p ->
  x \in orbit f p -> 
  ccw x (f x) q || ccw (f x) x q.
Proof.
move=> hulP xin.
have uni' : uniq inp by  move : uni; rewrite cons_uniq=> /andP[].
have : uniq [:: x; f x; q].
  have := uniq_hull2 s3 uni' xin hulP; rewrite -/f /= !inE !negb_or -andbA.
  move => /andP[] ->.
  by rewrite -!(eq_sym q) (@qnotiter x 0) // (@qnotiter x 1).
by move/Knuth_3/orP.
Qed.

Lemma sizeSpred (T : finType) (g : T -> T) x :
  (size (orbit g x)).-1.+1 = size (orbit g x).
Proof. by rewrite !size_orbit orderSpred. Qed.

Lemma size_orbit_stable :
  {in orbit f p, forall x, size (orbit f x) = size (orbit f p)}.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
move=> x xin; have [j ->]:= (orbit_rot_cycle cyc (orbit_uniq _ _) xin).
by rewrite size_rot.
Qed.

(* Important lemma, deserves a comment in any paper about this work. *)
Lemma sub_convex_triangle x i j :
 x \in orbit f p ->
 ((0 < i < j) && (j < size (orbit f p)))%N -> ccw x (iter i f x) (iter j f x).
Proof using s3 cot ccs.
move: (cot)=> [] _ [] _ [] incP [] hulP _ xin cmps.
have sos := size_orbit_stable.
have/iter_lt_order_neq/= xnj : (0 < j < fingraph.order f x)%N.
  rewrite -size_orbit sos //; case/andP: (cmps) => _ ->; rewrite andbT.
  by case/andP: cmps => /andP [h1 h2] _; rewrite (ltn_trans h1 h2).
have/iter_lt_order_neq/= xni : (0 < i < fingraph.order f x)%N.
  rewrite -size_orbit sos //; case/andP: (cmps) => /andP [_ h1] h2.
  by rewrite (ltn_trans h1 h2) andbT; case/andP: cmps => /andP [].
have itin : forall k, iter k f x \in inp.
  by move=> k; apply/(subsetP incP)/(subset_orbit xin)/mem_iter_orbit.
move: cmps (xnj); have [/eqP -> cmps _ | in1] := boolP (i == 1)%N.
  move: (cmps); rewrite -(sos _ xin).
  rewrite ltnSn andTb size_orbit=> /iter_lt_order_neq/= fxnj.
  by apply: (hulP _ xin _ (itin j)); rewrite eq_sym.
case/andP => /andP [i0 ij]; rewrite -(subnK ij) => js; move: (js).
move: (i0); rewrite leq_eqVlt eq_sym (negbTE in1) /= => igt1.
rewrite (subnK ij) in js.
have /iter_lt_order_neq fxni : (1 < i < fingraph.order f x)%N.
  by rewrite igt1 (ltn_trans ij) // -size_orbit sos.
elim: (j - i.+1)%N => [ | k IH].
  rewrite add0n => i1s xni1; apply/Knuth_1_1.
  by apply: (hulP _ (subset_orbit xin (mem_iter_orbit f x _)) _ (itin 0)).
rewrite !addnS !addSn in IH * => k2s xnk2.
have k1s : ((k + i).+1 < size (orbit f p))%N by apply: ltn_trans k2s.
have /iter_lt_order_neq xnk1 : (0 < (k + i).+1 < fingraph.order f x)%N.
  by rewrite -size_orbit sos // k1s ltn0Sn.
have /iter_lt_order_neq fxnk1 : (1 < (k + i).+1 < fingraph.order f x)%N.
  by rewrite -size_orbit sos // k1s -addSn (leq_trans igt1) // leq_addl.
have /iter_lt_order_neq fxnk2 : (1 < (k + i).+2 < fingraph.order f x)%N.
  by rewrite -size_orbit sos // k2s -!addSn (leq_trans igt1) // leq_addl.
have := fun n => hulP _ xin _ (itin n); rewrite -/f=> chx.
have xfxk1 : ccw x (f x) (iter (k + i).+1 f x) by apply: chx; rewrite eq_sym.
have xk1k2 : ccw x (iter (k + i).+1 f x) (iter (k + i).+2 f x).
  rewrite [iter _.+2 _ _]iterS; apply/Knuth_1_1/hulP=> //.
    by apply/(subset_orbit xin)/mem_iter_orbit.
  by apply/(subsetP incP).
by apply: (Knuth_5 _ xfxk1) => //; apply: IH || (apply:chx; rewrite eq_sym).
Qed.

Lemma pre_contiguous_red a b c:
  a \in orbit f p -> c \in orbit f p -> a != b -> f a != b ->
  ccw a (f a) b -> ccw a (f a) c -> ccw (f a) b c -> ccw (f a) b (f c).
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
move=> ao co anb fanb afab afac fabc.
have [/eqP -> | fcna] := boolP (f c == a); first by apply /Knuth_1.
have fanc : f a != c by move: afac=>/Knuth_1/oriented_diff.
have anc : a != c by move: afac=>/Knuth_1_1/oriented_diff; rewrite eq_sym.
have fanfc : f a != f c by rewrite (inj_in_eq convex_hull_path_inj ao co) anc.
have fain : f a \in inp by apply/(subsetP incP)/mem_orbit1.
have fcin : f c \in inp by apply/(subsetP incP)/mem_orbit1.
have facfc : ccw (f a) c (f c).
  by apply/Knuth_1_1/(hulP c co _ _ fanc fanfc).
have afafc : ccw a (f a) (f c) by apply: (hulP _ ao _ fcin); rewrite // eq_sym.
by apply: (Knuth_5' afab afac afafc fabc facfc).
Qed.

Lemma iter_size x : x \in orbit f p -> 
  iter (size (orbit f p)) f x = x.
Proof.
have ost : {in orbit f p, forall x, f x \in orbit f p}.
  by move=> z zin; apply: mem_orbit1.
move=> xin; rewrite -(size_orbit_stable xin) size_orbit.
by rewrite (iter_order_in ost convex_hull_path_inj).
Qed.

Lemma findexf x : x \in orbit f p ->
  findex f (f x) x = (size (orbit f p)).-1.
Proof.
move=> xo; move: (iter_size xo).
rewrite -(size_orbit_stable (mem_orbit1 xo)) -sizeSpred iterSr => xx.
have it : fconnect f (f x) x by rewrite -{2}xx fconnect_iter.
move/findex_max: (it); rewrite -size_orbit=> it'; move: (it').
rewrite -sizeSpred /= ltnS leq_eqVlt => /orP[/eqP // | abs].
move: (it); rewrite fconnect_orbit=> /(nth_index (f x)) => idxx.
have : findex f (f x) x < (fingraph.order f (f x)).-1 < fingraph.order f (f x).
  by rewrite -size_orbit abs sizeSpred leqnn.
move/iter_lt_order_neq; rewrite -size_orbit xx.
have := nth_traject f it' (f x); rewrite size_orbit -/(orbit f (f x)) idxx=> <-.
by rewrite eqxx.
Qed.

Lemma convex_hull_size_orbit_gt2 : 
  hull_prop inp m p -> 2 < size (orbit f p).
Proof.
move=> hulP.
have uni' : uniq inp by  move : uni; rewrite cons_uniq=> /andP[].
have := uniq_hull2 s3 uni' (orbit0 _ _) hulP.
rewrite size_orbit  /= !inE negb_or andbT -/f=> /andP[] /andP[] pf pff fpff.
rewrite /fingraph.order (cardD1 (f (f p))) (cardD1 (f p)) (cardD1 p).
by rewrite !inE (fconnect_iter f 2) fconnect1 connect0 pf pff fpff.
Qed.

Lemma contiguous_red a i j :
  a \in orbit f p -> 0 < i -> ccw a (f a) q -> ccw (iter i f a) (f a) q ->
  ccw (iter i f a) (iter i.+1 f a) q -> 
  i <= j <= size (orbit f p) -> ccw (iter j f a) (iter j.+1 f a) q.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
move=> ain i0 afaq ifafaq ii1q /andP[] ij; have sc:= sub_convex_triangle.
rewrite -(subnK ij); elim (j - i)=> [ | k IH] => // {j ij}.
rewrite !addSn; set j := _ .+1.
move: i0; rewrite leq_eqVlt => /orP [/eqP A | i1].
  by move/oriented_diff: ifafaq; rewrite -A eqxx.
have ij : (i < j)%N by rewrite /j ltnS leq_addl.
have j2 : (2 < j)%N by rewrite (leq_trans _ ij).
rewrite leq_eqVlt => /orP [/eqP -> | js].
  by rewrite //= iter_size.
have fain : f a \in orbit f p by apply/mem_orbit1.
have /(sc _ _ _ fain) : (0 < i.-1 < j.-1) && (j.-1 < size (orbit f p)).
  by to_lia; move: (leP js) (leP i1) (leP ij); ssr_lia.
rewrite -!iterSr (ltn_predK i1) (ltn_predK ij) => faij.
have/(sc _ _ _ ain) afai  : ((0 < 1 < i) && (i < size (orbit f p)))%N.
  by rewrite i1 !andTb (ltn_trans ij).
have/(sc _ _ _ ain) afaj : ((0 < 1 < j) && (j < size (orbit f p)))%N.
  by rewrite (ltn_trans i1) // size_orbit_stable.
have faqj : ccw (f a) q (iter j f a).
  by apply: (Knuth_5' _ afai) => //; apply/Knuth_1.
have/IH {IH} : k + i <= size (orbit f p).
  by rewrite (leq_trans _ js) // (leq_trans (leqnSn _)).
rewrite -[(k + i)%N]/j.-1 -[j.-1.+1]/j => jm1jp.
have jin : iter j f a \in orbit f p.
  by apply/(subset_orbit ain)/mem_iter_orbit.
have round : iter (size (orbit f p) - 1) f (iter j f a) = iter j.-1 f a.
  rewrite -iter_add [_ + j](_ : _ = j.-1 + size (orbit f p)).
    by rewrite iter_add iter_size.
  by move: (leP j2) (leP js); ssr_lia.
have/(sc _ _ _ jin) : (0 < 1 < size (orbit f p) - 1) &&
                       (size (orbit f p) - 1 < size (orbit f p)).
  by to_lia; move: (leP js) (leP j2); ssr_lia.
rewrite round -iter_add add1n=> /Knuth_1_1=> jm1jj1.
have jm2p1 : j.-2.+1 = j.-1 by move: (leP j2); ssr_lia.
have/(sc _ _ _ fain) : (0 < j.-2 < j.-1) && (j.-1 < size (orbit f p)).
  by to_lia; move: (leP j2) (leP js); ssr_lia.
rewrite -!iterSr (ltn_predK j2) jm2p1 => /Knuth_1 jm1jfa.
have/(sc _ _ _ fain)/Knuth_1 : (0 < j.-1 < j) && (j < size (orbit f p)).
  by to_lia; move: (leP j2) (leP js); ssr_lia.
rewrite -!iterSr (ltn_predK j2)=> jj1fa.
apply: (Knuth_5' _ jm1jfa) => //; last by apply/Knuth_1_1.
Qed.

Lemma red_path n x : x \in orbit f p -> 0 < n < size (orbit f p) ->
  (forall i, i < n -> ccw (iter i.+1 f x) (iter i f x) q) ->
    ccw (iter n f x) x q.
Proof.
move=> xin; elim: n => [| [_ | n IH]] // /andP [] n0 nor red.
  by apply:red.
have xn1n2 : ccw x (iter n.+1 f x) (iter n.+2 f x).
  have : (0 < n.+1 < n.+2) && (n.+2 < size (orbit f p)) by rewrite ltnSn.
  by apply: sub_convex_triangle.
apply: (Knuth_4 (Knuth_1 (Knuth_1 xn1n2))); apply/Knuth_1.
  apply/IH; first by to_lia; move: (leP nor); ssr_lia.
  by move=> i c; apply red; rewrite ltnW.
by apply/Knuth_1/red.
Qed.

Lemma blue_edge : 
  hull_prop inp m p -> exists2 x, (x \in orbit f p) & ccw x (f x) q.
Proof.
move=> hulP.
have uni' : uniq inp by move: uni; rewrite cons_uniq=> /andP[].
have [ | ] := boolP(~~[forall x, (x \in orbit f p) ==> ccw (f x) x q]).
  rewrite negb_forall=> /existsP[x]; rewrite negb_imply => /andP[] xin xblue.
  by exists x; have:= general_position_orbit hulP xin;
    rewrite ?(negbTE xblue) ?orbF.
rewrite negbK=>/forallP reds.
have sz : (0 < (size (orbit f p)).-1 < size (orbit f p))%N.
  by move: (convex_hull_size_orbit_gt2 hulP); case: (size _) => [ |[ | k]] //=.
have:= red_path (orbit0 f p) sz=> it; exists (iter (size (orbit f p)).-1 f p).
  by rewrite mem_iter_orbit.
rewrite -iterS sizeSpred size_orbit.
have := iter_order_in (fun y yin => mem_orbit1 yin) convex_hull_path_inj.
move=> ->; rewrite ?orbit0 // -size_orbit it // => e ein.
by move: (reds (iter e f p)); rewrite mem_iter_orbit implyTb /=.
Qed.

Lemma purple_point :
  exists a, [&& a \in orbit f p, ccw a (f a) q & ~~ ccw (f a) (f (f a)) q].
Proof using s3 cot uni redin redP ccs.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
have uni' : uniq inp by move: uni; rewrite cons_uniq=> /andP[].
move: (blue_edge hulP) => [y yin blue].
have [k oy] := orbit_rot_cycle cyc (orbit_uniq _ _) yin.
have : red_point \in orbit f y by rewrite oy mem_rot.
rewrite -fconnect_orbit => /iter_findex idred; move: redP; rewrite -idred.
elim: (findex _ _ _) => [ | n IH the_red].
  by rewrite /= => /Knuth_2; rewrite blue.
have nin : iter n f y \in orbit f p.
  by apply/(subset_orbit yin)/mem_iter_orbit.
have/orP[it|] := general_position_orbit hulP nin; last by apply: IH.
exists (iter n f y); rewrite -(same_orbit yin) mem_iter_orbit andTb.
  by move: it the_red; rewrite /= => -> /Knuth_2 ->.
Qed.

Lemma partition_red_blue:
  exists b, exists i,
  b \in orbit f p /\ 0 < i < size (orbit f p) /\
  (forall j, j < i -> ccw (iter j.+1 f b) (iter j f b) q) /\
  (forall j,  i <= j < size (orbit f p) -> ccw (iter j f b) (iter j.+1 f b) q).
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
have qnotin : q \notin inp by move:uni; rewrite cons_uniq=> /andP[].
have uni' : uniq inp by move:uni; rewrite cons_uniq=> /andP[].
have [b /andP[bin /andP [bblue fbnblue]]] := purple_point.
have chin i : iter i f b \in orbit f p.
  by apply/(subset_orbit bin)/mem_iter_orbit.
have fbo : f b \in orbit f p by apply/(mem_orbit1 bin).
move: fbnblue; have/orP[-> //| fbred _] := general_position_orbit hulP fbo.
have [j [reds [nextblue jb]]] : exists j,
     (forall k, 0 < k <= j -> ccw (iter k.+1 f b) (iter k f b) q) /\
     ccw (iter j.+1 f b) (iter j.+2 f b ) q /\ j < size (orbit f p).
  pose a v := ccw v (f v) q; set s := orbit f (f b).
  set j := find a s.
  have jb : j < size (orbit f (f b)).
    rewrite /j -has_find /a /s; apply/hasP; exists b =>//.
    by rewrite -[X in X \in _](iter_size bin) -sizeSpred iterSr mem_iter_orbit.
  exists j; split.
    move=> [ | k] //= kj; move: (kj)=> /(before_find (f b)); rewrite /a /s.
    rewrite /orbit -size_orbit nth_traject; last by apply: ltn_trans jb.
    rewrite -iterSr -!iterS => red.
  have := general_position_orbit hulP
            (subset_orbit bin (mem_iter_orbit _ b k.+1)).
    by rewrite red -iterS.
  move: (jb); rewrite /j /s -has_find => /(nth_find (f b)); rewrite /orbit.
  rewrite nth_traject -/j /a; last by rewrite -/j -size_orbit.
  rewrite -iterSr -iterS => ->; split=> //.
  by move: jb; rewrite size_orbit_stable.
exists (f b), j.
split; first by [].
have intj : 0 < j < size (orbit f p).
  rewrite jb andbT lt0n; apply/eqP=> j0; move: nextblue; rewrite j0 /=.
  by move => /Knuth_2; rewrite fbred.
split; first by [].
split; first by move=> k kj; rewrite -!iterSr; apply/reds; rewrite /=.
have frp : forall i, i < j -> ccw (iter i.+1 f (f b)) (iter i f (f b)) q.
  by move=> i ij; rewrite -!(iterSr _ f b); apply: reds.
have := red_path (mem_orbit1 bin) intj frp; rewrite -iterSr=> bigred.
have it := contiguous_red bin (isT : 0 < j.+1) bblue bigred nextblue.
by move=> k intk; rewrite -!iterSr; apply: it; rewrite ltnS.
Qed.

Section fixed_red_subpath.

Variables (b : P) (i : nat).

Hypothesis bin : b \in orbit f p.
Hypothesis ilto : 0 < i < size (orbit f p).
Let nf : P -> P := eval (new_map m b (iter i f b) q).
Hypothesis reds : forall k, k < i%N -> ccw (iter k.+1 f b) (iter k f b) q.
Hypothesis blues :
  forall k, (i <= k < size (orbit f p))%N ->
  ccw (iter k f b) (iter k.+1 f b) q.

Lemma iter_b_neq k l:
  k < l < size (orbit f p) -> iter k f b != iter l f b.
Proof.
by move=> ?; apply: iter_lt_order_neq; rewrite -size_orbit size_orbit_stable.
Qed.

Lemma igt0 : 0 < i.
Proof. by case/andP: ilto. Qed.

Lemma bred : ccw (f b) b q.
Proof. by rewrite -/(iter 0 f b) -iterS; apply/reds/igt0. Qed.

Lemma ilto' : i < size (orbit f p).
Proof. by case/andP: ilto. Qed.

Lemma qnb : q != b.
Proof. by apply: (qnotiter 0 bin). Qed.

Lemma nfb : nf b = q.
Proof. by rewrite /nf eval_update1. Qed.

Lemma nfq  : nf q = iter i f b.
Proof.
by rewrite /nf eval_update_dif ?eval_update1 // eq_sym (qnotiter 0 bin).
Qed.

Lemma nfk k : i <= k < size (orbit f p) -> nf (iter k f b) = iter k.+1 f b.
Proof.
move=> /andP[ik ko]; rewrite /nf eval_update_dif; last first.
  by rewrite -{1}/(iter 0 f b) iter_b_neq // ko (leq_trans _ ik) ?igt0.
by rewrite eval_update_dif ?qnotiter.
Qed.

Lemma new_path_offset k : 1 < k <= size (orbit f p) - i + 2 ->
  iter k nf b = iter (i + k - 2) f b.
Proof.
have ileo : i <= size (orbit f p) by rewrite leq_eqVlt orbC ilto'.
elim: k => [| [ | [ | k]] IH] //.
  by move=> _; rewrite addnK /= nfb nfq.
move=> /andP[] _; set k' := k.+2.
move=> klim; rewrite iterS IH; last first.
  by rewrite 2!ltnS leq0n; move: klim=>/ltnW.
move: klim; rewrite /k' !addnS !subSS !subn0 !ltnS addn0 => klim.
by rewrite nfk // leq_addr -ltn_subRL klim.
Qed.

Lemma new_path_cycle : iter (size (orbit f p) - i + 2) nf b = b.
Proof.
rewrite -[RHS](iter_size bin).
have ileo : i <= size (orbit f p) by rewrite leq_eqVlt orbC ilto'.
rewrite new_path_offset.
  by rewrite -(addnC 2) addnA (addnC i) -addnA addKn (addnC i) subnK.
by rewrite !addnS addn0 !ltnS leq0n leqnn.
Qed.

Lemma findex_eqns e :
  e \in orbit f p -> (findex f e (f e) = 1) /\ (findex f e (f (f e)) = 2).
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
have uni' : uniq inp by move: uni=> /andP[].
move=> eo; have := uniq_hull2 s3 uni' eo hulP.
have := convex_hull_size_orbit_gt2 hulP.
have <- : size (orbit f e) = size (orbit f p) by rewrite size_orbit_stable //.
rewrite size_orbit /findex /orbit.
case (fingraph.order _) => [ | [ | [ |k]]] //= _.
rewrite !inE !negb_or -!andbA -/f !eqxx.
by repeat (move=>/andP[] ne; rewrite ?(negbTE ne) {ne}).
Qed.

Definition purple (side : bool) := if side then iter i f b else b.

Lemma mem_purple_orbit e side : e \in orbit f p -> purple side \in orbit f e.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
move=> eo.
have [k oy] := orbit_rot_cycle cyc (orbit_uniq _ _) eo.
rewrite oy mem_rot; case: (side); last by apply: bin.
by apply/(subset_orbit bin)/mem_iter_orbit.    
Qed.

Lemma purpleP side : ccw (purple side) (f (purple side)) q = side.
Proof.
case:side.
  have /blues: i <= i < size (orbit f p).
    by rewrite leqnn ilto'.
  by [].
by rewrite /purple; apply/negbTE/Knuth_2/bred.
Qed.

Lemma purple_unique e side : e \in orbit f p -> ccw e (f e) q = ~~ side ->
  ccw (f e) (f (f e)) q = side -> f e = purple side.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
have [igt0 ilto'] := andP ilto.
move=> eo; have eob : e \in orbit f b.
  have [k oy] := orbit_rot_cycle cyc (orbit_uniq _ _) bin.
  by rewrite /f oy mem_rot.
move: eob; rewrite /orbit=> /trajectP=>[] []k. 
rewrite -size_orbit size_orbit_stable // => klto ek.
case: side; rewrite /purple.
  have [ | ] := boolP(k < i.-1).
    by rewrite -ltnS (prednK igt0) ek -!iterS => /reds/Knuth_2/negbTE ->.
  rewrite -leqNgt leq_eqVlt (prednK igt0)=> /orP[/eqP ik | ik ].
    by rewrite ek -iterS -ik (prednK igt0).
  have /blues : i <= k < size (orbit f p) by rewrite ik klto.
  by rewrite ek -iterS => ->.
rewrite ek /= -!iterS.
have [/reds/Knuth_2/negbTE -> //|] := boolP (k < i).
rewrite -leqNgt=> ik; move: klto; rewrite leq_eqVlt=> /orP[/eqP -> | k1lto].
  by rewrite iter_size.
have /blues -> // : i <= k.+1 < size (orbit f p).
by rewrite k1lto andbT; apply: (leq_trans ik).
Qed.

Lemma findex_decrease x y : x \in orbit f p -> y \in orbit f p ->
  x != y -> findex f x y = (findex f (f x) y).+1.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
move=> xo yo xny.
have yx : y \in orbit f x.
  have [k oy] := orbit_rot_cycle cyc (orbit_uniq _ _) xo.
  by rewrite /f oy mem_rot.
 move: (yx); rewrite /findex /orbit -!size_orbit.
rewrite !size_orbit_stable //; last by apply: mem_orbit1.
rewrite -sizeSpred trajectS trajectSr -cats1 inE eq_sym /=.
by rewrite (negbTE xny) orFb => yfx; rewrite index_cat yfx.
Qed.

Lemma find_next_purple_mapP side n e : e \in orbit f p ->
  ccw e (f e) q = ~~side -> findex f (f e) (purple side) <= n ->
  find_next_purple_map ccf q m side n (f e) = purple side.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
elim: n e => [ | n IH] e eo.
  rewrite leqn0 /= /findex=> cor sz.
  have /(nth_index e) := mem_purple_orbit side (mem_orbit1 eo).
  by rewrite (eqP sz) /orbit -orderSpred /=.
move=> ops guard; rewrite /=; case: ifP => [it | notit].
  by apply: purple_unique => //; apply/eqP.
apply: IH; first by apply/mem_orbit1.
  by move: notit; case: (ccw _ _ _); case side.
have feo : f e \in orbit f p by apply: mem_orbit1.
move: guard (mem_purple_orbit side feo).
have [/eqP abs _ | eneq iq ] := boolP (f e == purple side).
  by move: notit; rewrite -/f abs -/f purpleP; case: (side).
by move: iq; rewrite (findex_decrease feo (mem_purple_orbit side _)) // orbit0.
Qed.

Lemma map_sizeP' (x : P) : size (orbit (eval m) x) <= map_size m.
Proof.
have := orbit_uniq (eval m) x.
have /fpathP : exists n, behead (orbit (eval m) x) =
   traject (eval m) (eval m x) n.
  by rewrite /orbit -orderSpred trajectS /=; eexists; eauto.
rewrite /orbit -orderSpred trajectS [(behead _)]/=.
by apply: map_sizeP.
Qed.

Lemma find_purple_points_mapP : find_purple_points_map q m p = (b, iter i f b).
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
rewrite /find_purple_points_map.
have po : p \in orbit f p by apply: orbit0.
have fpo : f p \in orbit f p by apply: mem_orbit1.
have fisz x side : x \in orbit f p -> findex f x (purple side) <= map_size m.
  move=> xo; apply: leq_trans (map_sizeP' x); last first.
  have [k oy] := orbit_rot_cycle cyc (orbit_uniq _ _) xo.
  rewrite size_orbit -/f ltnW // findex_max //.
  by rewrite fconnect_orbit oy mem_rot mem_purple_orbit.
case: ifP => [blue_p | red_p].
  rewrite find_next_purple_mapP=> //=.
  rewrite find_next_purple_mapP=> //=; first by apply/negbTE/Knuth_2/bred.
    by apply/(fisz (f b) true)/mem_orbit1.
  by apply/(fisz (f p) false)/mem_orbit1/orbit0.
rewrite (find_next_purple_mapP po) //=.
  rewrite find_next_purple_mapP //=.
      by apply/(subset_orbit bin)/mem_iter_orbit.
    by apply: blues; rewrite leqnn ilto'.
  by apply: (fisz _ false); rewrite -iterS -(same_orbit bin) mem_iter_orbit.
apply/(fisz _ true)/mem_orbit1/orbit0.
Qed.

Lemma map_orbitP x n trailer :
  x \in orbit f p -> findex f (f x) b < n ->
  map_orbit (mf := mf) eq_op m x b n trailer =
  traject f x (findex f (f x) b).+1 ++ trailer.
Proof.
elim: n x => [| n IH] //= x xo.
rewrite ltnS.
case: ifP => [/eqP fxb | /negbT].
  have -> //: findex f (f x) b = 0.
  by rewrite /findex /orbit -orderSpred /= /f fxb eqxx.
rewrite -/f => fxnb; rewrite findex_decrease ?(mem_orbit1 xo) // => ifx.
by rewrite (IH _ (mem_orbit1 xo) ifx).
Qed.


Lemma traject_iota : traject f b (size (orbit f b)) =
      [seq iter k f b | k <- iota 0 (size (orbit f b))].
Proof.
suff : traject f (iter (size (orbit f b) - size (orbit f b)) f b)
        (size (orbit f b))
       =
       [seq iter k f b | k <- iota (size (orbit f b) - size (orbit f b))
                                   (size (orbit f b))].
  by rewrite subnn /=.
elim: {1 4 5 7 8}(size (orbit f b)) (leqnn (size (orbit f b)))
  => [// | l IHl /=].
by move=> llt; congr (_ :: _); rewrite -iterS subnSK //; apply/IHl/ltnW.
Qed.

Lemma im1red : ccw (iter i f b) (iter i.-1 f b) q.
Proof. by rewrite -{1}(prednK igt0); apply: reds; rewrite prednK ?igt0. Qed.

Lemma iblue : ccw (iter i.+1 f b) (iter i f b) q = false.
Proof. by apply/negbTE/Knuth_2/blues; rewrite leqnn ilto'.
Qed.

Lemma add_red_triangles_exact :
  add_red_triangles ccf q (map_orbit eq_op m b b (map_size m) [:: b]) tr =
  [seq [:: iter k.+1 f b; iter k f b; q] | k <- rev (iota 0 i)] ++ tr.
Proof.
rewrite map_orbitP // findexf // sizeSpred; last by rewrite map_sizeP'.
move: ilto => /andP[] igt0 ilto'.
have iblue : ccw (iter i.+1 f b) (iter i f b) q = false.
  by apply/negbTE/Knuth_2/blues; rewrite leqnn ilto'.
rewrite -(size_orbit_stable bin) traject_iota -(subn0 i).
move: (igt0); rewrite -(subnn (size (orbit f b))).
set sz := size (orbit f b).
elim: {1 4 6  7 9 11}sz (leqnn sz) tr => [ | l IH].
  by move=> _ tr'; rewrite !subn0 ltnNge /sz size_orbit_stable // (ltnW ilto').
move=> lsz tr'; rewrite leq_eqVlt=> /orP[ | lti].
  move=> qi; move: (qi).
  rewrite subnSK // => /eqP ati.
  have -> : sz - l.+1 = i.-1 by rewrite -(eqP qi).
  rewrite -(prednK igt0) /= subSnn /=.
  case h : l => [/= | l'].
    move: ati ilto'; rewrite h /sz size_orbit_stable //subn0 => <-.
    by rewrite ltnn.
  rewrite /= -iterS (prednK igt0) im1red /=.
  case h' : l' => [/= | l2].
    rewrite -{1}(iter_size bin) -(size_orbit_stable bin) -sizeSpred.
    by move: ati; rewrite h h' subn1 -/sz => ->; rewrite iblue.
  by rewrite /= -iterS iblue.
case h : l => [| l'].
  move: lti; rewrite h subn1 /sz sizeSpred.
  by rewrite size_orbit_stable // -/sz ltnNge (ltnW ilto').
set g := fun u => iter u f b.
rewrite [add_red_triangles _ _ _ _](_ : _ =
        if ccw (g (sz - l'.+1)) (g (sz - l'.+2)) q then
          add_red_triangles ccf q
           ([seq g i | i <- iota (sz - l'.+1) l'.+1] ++ [:: b])
           ([:: g (sz - l'.+1); g (sz - l'.+2); q]
           :: tr')
        else tr'); last first.
  by rewrite /= /g -iterS !subnSK // -?h //; apply: ltnW.
have step : sz - l = (sz - l.+1).+1 by rewrite subnSK.
have -> : sz - l'.+1 = (sz - l'.+2).+1 by rewrite -h.
  rewrite reds; last by rewrite -h (ltn_trans _ lti).
have szli : sz -l < i by move: lti; rewrite subnSK.
have llesz : l <= sz by apply: ltnW.
rewrite -h subnSK // IH // step /=.
have -> /= : i - (sz - l.+1) = (i - (sz - l.+1).+1).+1.
  by rewrite subnSK //; apply: ltn_trans szli; rewrite subnSK.
by rewrite rev_cons map_rcons -cats1 -catA.
Qed.

Lemma all_triangles_outtside_hull t :
  t \in add_red_triangles ccf q (map_orbit eq_op m b b (map_size m) [::b]) tr ->
  size t = 3 /\ uniq t /\ ccw_seq t.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
rewrite add_red_triangles_exact mem_cat => /orP[ | ?]; last by apply: trP.
  move/mapP=> []j; rewrite mem_rev mem_iota add0n leq0n andTb=> ji -> /=.
split; first by [].
split.
  rewrite !inE negb_or -!(eq_sym q) -!iterS !qnotiter // eq_sym !andbT.
  by apply: iter_b_neq; rewrite ltnSn (leq_trans _ ilto').
by apply: reds.
Qed.

Lemma ctr_outside_hull :
  flatten (add_red_triangles ccf q
        (map_orbit eq_op m b b (map_size m) [:: b]) tr)
    =i q :: inp.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
rewrite add_red_triangles_exact flatten_cat.
move=> z; apply/idP/idP; last first.
  rewrite inE=> /orP[/eqP -> | ].
    rewrite mem_cat; apply/orP; left; apply/flattenP.
    exists [:: iter 1 f b; iter 0 f b; q]; last by rewrite !subE.
    by apply/mapP; exists 0; rewrite // mem_rev mem_iota leqnn add0n igt0.
  by move=> zin; rewrite mem_cat ctr zin orbT.
rewrite mem_cat=> /orP[]; last by rewrite ctr inE orbC => ->.
by move=> /flattenP [t] /mapP[] j _ ->; rewrite !inE => /orP[|/orP[]]/eqP ->;
  rewrite ?eqxx // orbC (subsetP incP) // -(same_orbit bin) mem_iter_orbit.
Qed.

Lemma iter_injective_in (T : eqType) (g : T -> T) (S : pred T) (l : nat) :
  {in S, forall x, g x \in S} ->
  {in S &, injective g} -> {in S &, injective (iter l g)}.
Proof.
elim: l => [ | l IH] gst ginj x y xs ys; rewrite ?iterSr /=; first by [].
by move/(IH gst ginj (g x) (g y) (gst _ xs) (gst _ ys)); apply: ginj.
Qed.

Lemma f_stable : {in orbit f p, forall x, f x \in orbit f p}.
Proof. by move=> z zin; apply mem_orbit1. Qed.

Lemma sub_orbit_uniq k l l' : k + l <= size (orbit f p) ->
  uniq [seq iter (j + l') f b | j <- iota k l].
Proof.
move=> klo; apply/(uniqP p)=> x y; rewrite size_map !inE=> xbnd ybnd.
have [xbnd' ybnd'] : x < l /\ y < l.
    by move: xbnd ybnd; rewrite size_iota; split.
have [kxbnd kybnd] : k + x < size (orbit f p) /\ k + y < size (orbit f p).
  by split; apply: leq_trans klo; rewrite ltn_add2l.
have [kxbnd' kybnd'] : k + x < size (orbit f b) /\ k + y < size (orbit f b).
  by rewrite size_orbit_stable // split.
rewrite !(nth_map 0) // !nth_iota // -!(addnC (l')). 
rewrite [LHS]iter_add [RHS]iter_add => qq.
have/(uniqP p):= orbit_uniq f b=> /(_ (k + x) (k + y)).
rewrite size_orbit_stable // !inE kxbnd kybnd !nth_orbit //.
move/(iter_injective_in f_stable convex_hull_path_inj): (qq).
rewrite -!(same_orbit bin) !mem_iter_orbit=>/(_ isT isT)=> kxky.
move=>/(_ isT isT kxky) main.
by apply/eqP; rewrite -(eqn_add2l k); apply/eqP/main.
Qed.

Lemma cycle_orbit_nf_val :
  fcycle nf 
    (q :: [seq iter (k + i) f b | k <- iota 0 (size (orbit f p) - i).+1]).
Proof.
apply/(pathP p)=>[] [| k]; first by move=>_ /=; rewrite nfq eqxx.
rewrite size_rcons size_map size_iota 1!ltnS leq_eqVlt=>/orP[/eqP-> |].
  rewrite nth_rcons size_map size_iota ltnn eqxx.
  have szi : 0 < size (orbit f p) - i by rewrite subn_gt0 ilto'.
  set s' := rcons _ _; rewrite /= nth_rcons size_map size_iota leqnn.
  rewrite (nth_map 0); last by rewrite size_iota leqnn.
  rewrite nth_iota // add0n subnK; last by rewrite (ltnW ilto').
  by rewrite iter_size // nfb.
set s' := rcons _ _; rewrite ltnS /= => kbnd; rewrite nth_rcons.
rewrite size_map size_iota kbnd (nth_map 0); last by rewrite size_iota.
rewrite nth_iota // add1n addSn /s' nth_rcons size_map size_iota ltnS.
have kbnd' := ltnW kbnd.
rewrite (ltnW kbnd) (nth_map 0); last by rewrite size_iota.
rewrite nth_iota // add0n; apply/eqP/nfk.
by rewrite leq_addl addnC -ltn_subRL kbnd.
Qed.

Lemma orbit_nf : orbit nf q =
   q :: [seq iter (k + i) f b | k <- iota 0 (size (orbit f p) - i).+1].
Proof.
set s := _ :: map _ _.
have uns : uniq s.
  rewrite /s cons_uniq; apply/andP; split.
    by apply/negP=> /mapP [k _] /eqP; apply/negP/qnotiter/bin.
  apply sub_orbit_uniq; rewrite add0n -{2}(subn0 (size (orbit f p))).
  by rewrite ltn_sub2l ?igt0 // -sizeSpred.
have := cycle_orbit_nf_val; rewrite -/s=> cycs.
have qs : q \in s by rewrite !subE.
have := order_cycle cycs uns qs=> oq.
apply/esym; rewrite /orbit.
apply: (eq_from_nth (x0 := p)); first by rewrite size_traject.
move=> [| k]; first by rewrite /s /= oq /s.
move=> kbnd; rewrite oq [RHS](set_nth_default q) ?size_traject //.
rewrite nth_traject // -nfb -iterSr new_path_offset; last first.
  rewrite /= !addnS addn0 (leq_trans kbnd) // /s /= size_map size_iota.
  by rewrite leqnn.
rewrite /s; set s' := map _ _; rewrite !addnS !subSS subn0 /s /=.
rewrite (nth_map 0); last first.
  by move: kbnd; rewrite /s /= size_map size_iota ltnS.
rewrite nth_iota; last first. 
  by move: kbnd; rewrite /s /= size_map size_iota ltnS.
by rewrite add0n addnC.
Qed.

Lemma orbit_nf_inc : orbit nf q \subset q :: inp.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
rewrite orbit_nf; apply/subsetP=> z; rewrite inE => /orP[/eqP ->|].
  by rewrite !subE.
move/mapP=> [j] _ ->; rewrite inE; apply/orP; right; apply: (subsetP incP).
by rewrite -(same_orbit bin) mem_iter_orbit.
Qed.

Lemma new_hulP :
  forall r, r \in orbit nf q ->
  forall x, x \in q :: inp -> x != r -> x != nf r -> ccw r (nf r) x.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
have uni' : uniq inp by move: uni=> /andP[].
have iterin k : iter k f b \in orbit f p.
  by rewrite -(same_orbit bin) mem_iter_orbit.
move=> r; rewrite orbit_nf inE => /orP [/eqP -> | ].
  move=> x; rewrite !inE=> /orP[/eqP -> |]; first by rewrite eqxx.
  move=> xinp xnq xnfq; have [/eqP -> | xnim1] := boolP(x == iter i.-1 f b).
    have/reds : i.-1 < i by rewrite (ltn_predK igt0) leqnn.
    by move/Knuth_1_1; rewrite (ltn_predK igt0) nfq.
  have [/eqP ->| xnnf2q ] := boolP (x == nf (nf q)).
    apply/Knuth_1_1; rewrite nfq nfk.
      by apply: blues; rewrite leqnn ilto'.
    by rewrite leqnn ilto'.
  have nf2v : nf (nf q) = iter i.+1 f b by rewrite nfq nfk ?leqnn ?ilto'.
  (* preparing the use of Knuth_5 *)
  have im1o : iter i.-1 f b \in orbit f p by rewrite iterin.
  have := uniq_hull2 s3 uni' im1o hulP=> /=; rewrite !inE negb_or -!iterS -/f.
  rewrite (prednK igt0)=> /andP[] /andP[] dif1 dif2 /andP[] dif3 _.
  have ff2x : ccw (nf q) (nf (nf q)) x.
    rewrite nf2v nfq //; apply/hulP => //; first by rewrite -nfq.
    by rewrite -iterS -nf2v.
  have ff2im1 : ccw (nf q) (nf (nf q)) (iter i.-1 f b).
    rewrite nf2v nfq //; apply/hulP => //.
    by apply/(subsetP incP).
  have ff2q : ccw (nf q) (nf (nf q)) q.
    by rewrite nf2v nfq; apply/blues; rewrite leqnn ilto'.
  have fxim1 : ccw (nf q) x (iter i.-1 f b).
    rewrite nfq; apply/Knuth_1; rewrite -{2}(prednK igt0); apply: hulP=> //.
    by rewrite -iterS (prednK igt0) -nfq.
  have fqim1q : ccw (nf q) (iter i.-1 f b) q.
    by rewrite nfq -{1}(prednK igt0); apply: reds; rewrite (prednK igt0).
  by apply/Knuth_1_1/(@Knuth_5 (nf q) (nf (nf q)) x (iter i.-1 f b) q).
case/mapP=> [j]; rewrite mem_iota add0n ltnS leq0n andTb leq_eqVlt.
move=> /orP[/eqP -> | jlt].
  rewrite (subnK (ltnW ilto')) iter_size ?bin // => ->.
  move=> x; rewrite !inE=> /orP[/eqP -> |]; first by rewrite nfb eqxx.
  move=> xinp xnq xnfq; have [/eqP -> | xnim1] := boolP(x == f b).
    by have/reds := igt0; rewrite nfb /=; move/Knuth_1.
  rewrite nfb.
  have [/eqP ->| ] := boolP (x == iter (size (orbit f p)).-1 f b).
    set b' := iter _ _ _.
    have nfb' : f b' = b by rewrite /b' -iterS sizeSpred iter_size.
    apply/Knuth_1; rewrite -{1}nfb' /b' -iterS; apply: blues.
    by rewrite -ltnS sizeSpred leqnn ilto'.
  set b' := iter _ _ _ => xnb'.
  have b'o : b' \in orbit f p by rewrite /b' iterin.
  (* preparing the use of Knuth_5' *)
  have := uniq_hull2 s3 uni' b'o hulP=> /=; rewrite !inE negb_or -!iterS -/f.
  rewrite sizeSpred /= (iter_size bin).
  move => /andP[] /andP[] dif1 dif2 /andP[] dif3 _.
  have fb' : f b' = b by rewrite -(iter_size bin) /b' -sizeSpred /=.
  have b'bq : ccw b' b q.  
    by rewrite -fb' /b' -iterS; apply: blues; rewrite -ltnS sizeSpred ilto' /=.
  have b'bfb : ccw b' b (f b).
    rewrite -fb'; apply/hulP=> //.
        by apply/(subsetP incP); rewrite -/b' -!iterS iterin.
      by rewrite eq_sym /b' -!iterS sizeSpred /= (iter_size bin).
    by rewrite -/f fb' eq_sym.
  have b'bx : ccw b' b x by rewrite -fb'; apply/hulP=> //; rewrite -/f fb'.
  have bqfb : ccw b q (f b).
    by apply/Knuth_1; rewrite -/(iter 0 f b) -iterS; apply/reds/igt0.
  have bfbx : ccw b (f b) x by apply/hulP.
  by apply/(@Knuth_5' b' b q (f b) x).
have jiint : i <= j + i < size (orbit f p).
  by rewrite leq_addl addnC -ltn_subRL jlt.
move=> -> x; rewrite inE nfk // => /orP [/eqP -> _ _ | ]; first by apply: blues.
by apply/hulP.
Qed.

Lemma no_i_inside_outside t1 t2 p1:
  t1 \in tr ->
  t2 \in [seq [:: iter k.+1 f b; iter k f b; q] | k <- rev (iota 0 i)] ->
  inside_triangle_seq t1 p1 ->
  ~inside_triangle_seq t2 p1.
Proof.
have uni' : uniq inp by move: uni; rewrite /= => /andP[].
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
have iterin k : iter k f b \in orbit f p.
  by rewrite -(same_orbit bin) mem_iter_orbit.
move=> t1tr /mapP[] k; rewrite mem_rev mem_iota add0n leq0n andTb.
move=> ki t2q p1t1 p1t2.
have kin : iter k f b \in orbit f p by rewrite iterin.
move: (p1t2); rewrite t2q=> /andP[] /andP[]/Knuth_1/oriented_diff.
rewrite eq_sym=> p1nk _ /Knuth_1/oriented_diff; rewrite eq_sym => p1nk1.
have /Knuth_2/negbTE kk1p1 := inside_triangle_in_hull
          s3 (perm_to_rem t1tr) uni' trP ctr hulP kin p1t1 p1nk p1nk1.
by move:p1t2; rewrite t2q /= kk1p1.
Qed.

Lemma pre_no_i_outside k l p1 :
  k + l < i ->
  inside_triangle_seq [:: iter k.+1 f b; iter k f b; q] p1 ->
  ccw q (iter (k + l).+1 f b) p1.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
have iterin u : iter u f b \in orbit f p.
  by rewrite -(same_orbit bin) mem_iter_orbit.
have iinp u : iter u f b \in q :: inp.
  by rewrite inE; apply/orP; right; apply/(subsetP incP)/iterin.
have binn : b \in orbit nf q.
  rewrite orbit_nf inE; apply/orP; right; apply/mapP.
  exists (size (orbit f p) - i).
    by rewrite mem_iota leq0n add0n leqnn.
  by rewrite (subnK (ltnW ilto')) (iter_size bin).
move=> kli p1k; move: (p1k) => /andP[] /andP[] k1kp1 kqp1 qkp1.
have [/eqP -> | p1nb] := boolP(p1 == b).
  apply/Knuth_1; rewrite -nfb; apply/new_hulP=> //.
    rewrite eq_sym -{1}/(iter 0 f b); apply: iter_b_neq.
    by rewrite ltnS leq0n (leq_trans _ ilto').
  by rewrite nfb eq_sym qnotiter.
elim: l kli => [|l IH]; first by rewrite addn0.
rewrite addnS=> kli; have /IH IH' := ltnW kli.
have kl1nb : iter (k + l).+1 f b != b.
  rewrite eq_sym -{1}/(iter 0 f b); apply iter_b_neq.
  by rewrite ltnS leq0n (ltn_trans _ ilto').
have kl1nq : iter (k + l).+1 f b != nf b.
  by rewrite eq_sym nfb qnotiter.
have kl2nb : iter (k + l).+2 f b != b.
  rewrite eq_sym -{1}/(iter 0 f b); apply iter_b_neq.
  by rewrite ltnS leq0n (leq_trans _ ilto').
have kl2nq : iter (k + l).+2 f b != nf b.
  by rewrite eq_sym nfb qnotiter.
have := (new_hulP binn (iinp (k + l).+1) kl1nb kl1nq); rewrite nfb=> bqkl1.
have := (new_hulP binn (iinp (k + l).+2) kl2nb kl2nq); rewrite nfb=> bqkl2.
have bqp1 : ccw b q p1.
  have s3' : 2 < size (q :: inp) by rewrite /= ltnW.
  have p1nq : p1 != nf b by move/Knuth_1_1/oriented_diff: IH'; rewrite nfb.
  have/perm_to_rem prt1 : [:: iter k.+1 f b; iter k f b; q] \in 
       add_red_triangles ccf q (map_orbit eq_op m b b (map_size m) [:: b]) tr.
    rewrite add_red_triangles_exact mem_cat;apply/orP; left; apply/mapP.
    exists k => //; rewrite mem_rev mem_iota leq0n add0n (leq_trans _ kli) //.
    by rewrite ltnS -!addnS leq_addr.
  by have := inside_triangle_in_hull  s3' prt1 uni all_triangles_outtside_hull
   ctr_outside_hull new_hulP binn p1k p1nb p1nq; rewrite -/nf nfb.
apply: (@Knuth_5' b q (iter (k + l).+2 f b) (iter (k + l).+1 f b) p1)=> //.
by apply/Knuth_1_1/reds.
Qed.

Lemma no_i_outside_hull t1 t2 p1 :
  t1 \in add_red_triangles ccf q
        (map_orbit eq_op m b b (map_size m) [:: b]) tr ->
  t2 \in add_red_triangles ccf q
        (map_orbit eq_op m b b (map_size m) [:: b]) tr ->
  inside_triangle_seq t1 p1 ->
  inside_triangle_seq t2 p1 -> t1 = t2.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
rewrite add_red_triangles_exact; set ts := map _ _.
rewrite 2!mem_cat=> /orP[t1ts | t1tr]; last first.
  move=>/orP [t2ts | t2tr]; last by apply: no_i.
  by move=> p1t1 p1t2; have := no_i_inside_outside t1tr t2ts p1t1 p1t2.
move=>/orP [t2ts | t2tr]; last first.
  by move=> p1t1 p1t2; have := no_i_inside_outside t2tr t1ts p1t2 p1t1.
move: t1ts t2ts; rewrite /ts=>/mapP[k kin ->] /mapP[l lin ->].
wlog : k l kin lin / k <= l.
move=> main; have[klel | ] := boolP(k <= l); first by apply: main.
  by rewrite -ltnNge=>/ltnW cmp ? ?; apply/esym; apply: (main _ _ _ _ cmp).
rewrite leq_eqVlt=> /orP[/eqP -> // | kltl].
set l' := (l - k.+1); have lq : l = k.+1 + l' by rewrite -(subnK kltl) addnC.
move=> p1t1.
have/pre_no_i_outside : k + l' < i.
  move: lin; rewrite mem_rev mem_iota leq0n add0n andTb=> lin.
  by rewrite (leq_trans _ lin) // ltnS lq leq_add2r.
move=> /(_ _ p1t1); rewrite -addSn -lq=> abs /andP[] /andP[] l1lp1 /Knuth_2.
by rewrite abs.
Qed.

Lemma convex_hull_one_triangle x k :
  hull_prop inp m p ->
  orbit f p \subset inp ->
  k.+1 < size (orbit f p) -> ccw (iter k.+1 f b) b x &&
   ccw b (iter k f b) x && ccw (iter k f b) (iter k.+1 f b) x ->
  (forall k' : nat,
   k' < k.+1 -> ccw (iter k' f b) (iter k'.+1 f b) x).
Proof.
move=> hulP incP kp1lto /andP [] /andP [] kp1bx bkx kkp1x k'.
have klto : (k < size (orbit f p)) by apply: (ltn_trans (ltnSn _)).
rewrite ltnS leq_eqVlt => /orP [/eqP -> // | k'lt].
have [/eqP k0 | kn0] := boolP(k == 0)%N; first by move: k'lt; rewrite k0 ltn0.
have kgt0 : (0 < k)%N by rewrite lt0n.
have bnk : b != iter k f b.
  by have/iter_b_neq : 0 < k < size (orbit f p) by rewrite lt0n kn0.
have iterin u : iter u f b \in orbit f p.
  by rewrite -(same_orbit bin) mem_iter_orbit.
have kkp1b : ccw (iter k f b) (iter k.+1 f b) b.
  have:= (hulP _ (iterin k) _ (subsetP incP _ bin) bnk); rewrite -iterS -/f.
  have/iter_b_neq -> : 0 < k.+1 < size (orbit f p) by rewrite ltn0Sn.
  by apply.
have bnkp1 : b != iter k.+1 f b.
  by have/iter_b_neq-> // : 0 < k.+1 < size (orbit f p) by rewrite ltn0Sn.
have := hulP _ (iterin k'); rewrite -iterS=> chk'.
have k'lto : k' < size (orbit f p) by apply/(ltn_trans k'lt).
move: kkp1b=> /Knuth_1=> kp1bk.
have/Knuth_1_1 kp1bfb : ccw b (f b) (iter k.+1 f b).
  move: (hulP _ bin (iter k.+1 f b)); rewrite (subsetP incP) //.
  rewrite -(eq_sym b) bnkp1 -(eq_sym (f b)).
  have/iter_b_neq -> : 1 < k.+1 < size (orbit f p) by rewrite ltnS lt0n kn0.
  by apply.
have [/eqP k1 | kn1] := boolP (k == 1%N).
  by move: k'lt bkx; rewrite k1 ltnS leqn0=> /eqP ->.
have bfbk : ccw b (f b) (iter k f b).
  move: (hulP _ bin (iter k f b)); rewrite (subsetP incP) //.
  rewrite !(eq_sym (iter k f b)).
  have/iter_b_neq -> : 0 < k < size (orbit f p) by rewrite kgt0.
  have/iter_b_neq -> : 1 < k < size (orbit f p).
    by rewrite ltn_neqAle eq_sym kn1 kgt0.
  by apply.
have bfbx : ccw b (f b) x by apply: (Knuth_5' kp1bfb kp1bk kp1bx bfbk bkx).
have binp : b \in inp by apply: (subsetP incP).
have [/eqP -> // | ] := boolP(k' == 0%N);rewrite -lt0n => k'gt0.
have [k'p1k | gen ] := boolP (k'.+1 == k).
  move/Knuth_1_1: kp1bk => kkp1b.
  have kbk' : ccw (iter k f b) b (iter k' f b).
    apply/Knuth_1; rewrite -(eqP k'p1k); move: (hulP _ (iterin k') _ binp).
    have/iter_b_neq -> : 0 < k' < size (orbit f p).
      by rewrite  k'gt0 (ltn_trans k'lt).
    have/iter_b_neq -> : 0 < k'.+1 < size (orbit f p).
      by rewrite ltn0Sn (eqP k'p1k).
    by apply.
  have kkp1k': ccw (iter k f b) (iter k.+1 f b) (iter k' f b).
    move: (hulP _ (iterin k) (iter k' f b)); rewrite (subsetP incP) //.
    have/iter_b_neq -> // : (k' < k.+1 < size (orbit f p))%N.
      by rewrite (eqP k'p1k) leqnSn.
    have/iter_b_neq -> : (k' < k < size (orbit f p))%N.
      by rewrite (eqP k'p1k) leqnn.
    by apply.
  have/Knuth_1_1 := Knuth_5 kkp1x kkp1b kkp1k' (Knuth_1 bkx) kbk'.
  by rewrite -(eqP k'p1k).
move: kp1bx kkp1x=> /Knuth_1 bxkp1 /Knuth_1_1 xkkp1.
have k'kb : ccw (iter k' f b) (iter k f b) b.
  have trk : (0 < k' < k) && (k < size (orbit f p)) by rewrite k'gt0 k'lt.
  by have/Knuth_1 := sub_convex_triangle bin trk.
have k'p1lt : k'.+1 < k by move: k'lt; rewrite leq_eqVlt (negbTE gen).
have k'k'p1b : ccw (iter k' f b) (iter k'.+1 f b) b.
  move: (chk' b); rewrite binp.
  have/iter_b_neq -> : 0 < k' < size (orbit f p).
    by rewrite k'gt0 (ltn_trans k'lt). 
  have/iter_b_neq -> : 0 < k'.+1 < size (orbit f p).
    by rewrite ltn0Sn (ltn_trans k'p1lt) // (ltn_trans ki).
  by apply.
have k'k'p1k : ccw (iter k' f b) (iter k'.+1 f b) (iter k f b).
  move: (chk' (iter k f b)); rewrite (subsetP incP) // !(eq_sym (iter k f b)).
  have/iter_b_neq -> : k' < k < size (orbit f p) by rewrite k'lt.
  have/iter_b_neq -> : k'.+1 < k < size (orbit f p) by rewrite k'p1lt.
  by apply.
have k'k'p1kp1 : ccw (iter k' f b) (iter k'.+1 f b) (iter k.+1 f b).
  move: (chk' (iter k.+1 f b)); rewrite (subsetP incP) //
         !(eq_sym (iter k.+1 f b)).
  have/iter_b_neq -> : k' < k.+1 < size (orbit f p).
    by rewrite (ltn_trans _ (ltnSn k)).
  have/iter_b_neq -> : k'.+1 < k.+1 < size (orbit f p).
    by rewrite (ltn_trans k'p1lt (ltnSn _)).
  by apply.
by apply:(inside_triangle_left_of_edge k'k'p1b k'k'p1k k'k'p1kp1 bkx).
Qed.

Lemma convex_hull_split x k :
  k < size (orbit f p) ->
  ccw b (iter k f b) x ->
  (forall u, x != iter u f b) ->
  (forall l, k <= l < size (orbit f p) -> ccw (iter l f b) (iter l.+1 f b) x) ->
  forall l, l < k -> ccw (iter l f b) (iter l.+1 f b) x.
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
move=> klto bkx xniter blues_k.
suff main : forall j, j + k < size (orbit f p) ->
  ccw (iter (j + k) f b) b x ->
  forall l : nat, l < k -> ccw (iter l f b) (iter l.+1 f b) x.
  have arith : ((size (orbit f p)).-1 - k + k = (size (orbit f p)).-1)%N.
    by rewrite subnK // -ltnS sizeSpred.
  apply: (main ((size (orbit f p)).-1 - k)%N) => //.
    by rewrite arith sizeSpred.
  rewrite -2![in X in ccw _ X _](iter_size bin, sizeSpred) subnK.
    by apply:blues_k; rewrite -ltnS sizeSpred klto leqnn.
  by rewrite -ltnS sizeSpred.
elim => [ | j IH] kp1lto lbx l llek.
  by move/Knuth_2: lbx; rewrite add0n bkx.
have jklto : j + k < size (orbit f p).
  by apply: ltn_trans kp1lto; apply: ltnSn.
move: kp1lto; rewrite addSn => kp1lto.
have/Knuth_3[bjx |jbx] : uniq [:: iter 0 f b; iter (j + k) f b; x].
  have kn0 : k != 0%N by apply/negP=> /eqP A; move: llek; rewrite A ltn0.
  have/iter_b_neq : (0 < j + k < size (orbit f p))%N.
    by rewrite jklto; move: kn0; case: (k)=> // n _; rewrite addnS ltn0Sn.
  by rewrite !cons_uniq !inE -!(eq_sym x) negb_or !xniter => ->.
  have := convex_hull_one_triangle hulP incP kp1lto=> /(_ x).
  rewrite bjx lbx blues_k; last by rewrite leq_addl jklto.
  move=> /(_ isT); apply.
  by apply: (leq_trans llek); rewrite -addSn leq_addl.
by apply: IH.
Qed.

Lemma cover_hull_outside_hull r :
  inh (eval (new_map m b (iter i f b) q)) q r ->
  r \in q :: inp \/
  (exists t : seq_eqType P,
     t \in add_red_triangles ccf q
             (map_orbit eq_op m b b (map_size m) [:: b]) tr /\
     inside_triangle_seq t r).
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
rewrite /inh -/nf=> xh.
set ts' := add_red_triangles _ _ _ _.
have b_new_orbit : b \in orbit nf q.
  rewrite orbit_nf inE; apply/orP/or_intror/mapP; exists (size (orbit f p) - i).
    by rewrite mem_iota leq0n /=.
  by rewrite (subnK (ltnW ilto')) iter_size.
have [| rnotin] := boolP(r \in q :: inp);[left => //| right].
have rniter : forall k, r != iter k f b.
  move=> k; apply/eqP=> abs; case/negP: rnotin.
  rewrite inE;apply/orP/or_intror/(subsetP incP); rewrite abs.
  by rewrite -(same_orbit bin) mem_iter_orbit.
have nfi : nf (iter i f b) = (iter i.+1 f b).
  by apply: nfk; rewrite leqnn ilto'.
have blues_r k (kint : i <= k < size (orbit f p)):
     ccw (iter k f b) (iter k.+1 f b) r.
  suff/xh /= : iter k f b \in orbit nf q.
    by rewrite nfk.
  rewrite orbit_nf inE; apply/orP/or_intror/mapP/(ex_intro2 _ _ (k - i)).
    rewrite mem_iota leq0n andTb add0n ltnS leq_sub2r //.
    by move: kint=> /andP[] _ /ltnW.
  by rewrite subnK; move: kint=> /andP[].
have kqfk k : k < i -> ccw (iter k f b) q (iter k.+1 f b).
  by move=> klti; apply/Knuth_1/reds.
have main: forall k, k <= i -> inside_triangle_seq [:: b; q; iter k f b] r ->
   (exists t, t \in ts' /\ inside_triangle_seq t r) \/
   (forall k', k' < k -> ccw (iter k' f b) (iter k'.+1 f b) r).
  elim=>[ | k IH]; first by move=> _ _ ; right.
  have [/eqP -> ki insx| kn0 ki xins] := boolP(k == 0).
    left; exists [:: f b; b; q].
    split; last by rewrite inside_triangle_seq_K1.
    rewrite /ts' add_red_triangles_exact mem_cat; apply/orP/or_introl/mapP.
    by exists 0 => //; rewrite mem_rev mem_iota leq0n andTb.
(* preparatory work for choose_sub_triangle. *)
  have kgt0 : 0 < k by rewrite lt0n.
  have bnk : b != iter k f b.
    have/iter_b_neq// : 0 < k < size (orbit f p).
      by rewrite lt0n kn0 (ltn_trans ki) // ilto'.
  have bnkp1 : b != iter k.+1 f b.
    have/iter_b_neq -> // : 0 < k.+1 < size (orbit f p).
    by rewrite ltn0Sn (leq_ltn_trans ki) // ilto'.
  have knkp1 : iter k f b != iter k.+1 f b.
    have/iter_b_neq// : k < k.+1 < size (orbit f p).
    by rewrite ltnSn (leq_ltn_trans ki) // ilto'.
  have bpk : ccw b q (iter k f b).
    have := new_hulP b_new_orbit; rewrite nfb=> /(_ (iter k f b)).
    rewrite !(eq_sym (iter k f b)) bnk qnotiter //.
    rewrite inE (subsetP incP) ?orbT; first by apply.
    by rewrite -(same_orbit bin) mem_iter_orbit.
  have bpkp1 : ccw b q (iter k.+1 f b).
    have := new_hulP b_new_orbit; rewrite nfb=> /(_ (iter k.+1 f b)).
    rewrite !(eq_sym (iter k.+1 f b)) bnkp1 qnotiter //.
    rewrite inE (subsetP incP) ?orbT; first by apply.
    by rewrite -(same_orbit bin) mem_iter_orbit.
  have kkp1b : ccw (iter k f b) (iter k.+1 f b) b.
    apply: hulP=> //; first by rewrite -(same_orbit bin) mem_iter_orbit.
    by apply/(subsetP incP).
  have kins : inside_triangle_seq [:: b; q; iter k.+1 f b] (iter k f b).
    by rewrite /inside_triangle_seq bpk (Knuth_1 kkp1b) (Knuth_1_1 (reds ki)).
    move: (rniter k); rewrite eq_sym=> knr.
    have := choose_sub_triangle knr kins xins; rewrite /inside_triangles.
  move=>[] t; rewrite !inE=>[][]/orP[| /orP []]/eqP -> {t}.
      rewrite inside_triangle_seq_K1=> kbqr.
      case: (IH (ltnW ki) kbqr) => [ | inpath]; first by left.
      suff kkp1r : ccw (iter k f b) (iter k.+1 f b) r.
        right=> k'; rewrite leq_eqVlt=>/orP[k'kp1 | k'kp1].
          by move: k'kp1; rewrite eqSS=> /eqP ->.
        by move: k'kp1; rewrite ltnS; apply: inpath.
      have chm1 s :
        [&& (s \in inp), (s != iter k.-1 f b) & (s  != iter k f b)] ->
        ccw (iter k.-1 f b) (iter k f b) s.
        move=> /andP[] sinp /andP[] snkm1 snk.
        move: (hulP (iter k.-1 f b)); rewrite -(same_orbit bin) mem_iter_orbit.
        move=> /(_ isT _ sinp snkm1); rewrite -iterS (prednK kgt0) snk.
        by apply.
      have km1kkp1 : ccw (iter k.-1 f b) (iter k f b) (iter k.+1 f b).
        apply: chm1.
        rewrite (subsetP incP); last first.
          by rewrite -(same_orbit bin) mem_iter_orbit.
        rewrite andTb !(eq_sym (iter k.+1 f b)).
        have /iter_b_neq -> : k.-1 < k.+1 < size (orbit f p).
          rewrite (leq_ltn_trans ki ilto') andbT (ltn_trans _ (ltnSn _)) //. 
          by rewrite prednK ?leqnn.
        have /iter_b_neq -> // : k < k.+1 < size (orbit f p).
        by rewrite (leq_ltn_trans ki ilto') // ltnSn.
      have km1kr : ccw (iter k.-1 f b) (iter k f b) r.
        rewrite -[in X in ccw _ X _](prednK kgt0); apply: inpath.
        by rewrite prednK ?leqnn.
      have km1kb : ccw (iter k.-1 f b) (iter k f b) b.
        apply: chm1; rewrite (subsetP incP) // bnk andTb andbT.
        apply/negP=> /eqP A.
        by move: kbqr=>/=; rewrite {4}A (negbTE (Knuth_2 km1kr)) andbF. 
      by apply: (Knuth_5' km1kkp1 km1kb) => //; move: kbqr => /andP[].
    move=> it; left; exists [:: iter k.+1 f b; iter k f b; q].
    rewrite inside_triangle_seq_K1 it /ts' add_red_triangles_exact; split=> //.
    rewrite mem_cat; apply/orP/or_introl/mapP/(ex_intro2 _ _ k) => //.
    by rewrite mem_rev mem_iota leq0n add0n ki.
  rewrite inside_triangle_seq_K1 /inside_triangle_seq=> kp1bk ; right.
  have := convex_hull_one_triangle hulP incP=> /(_ r k).
  by rewrite (leq_ltn_trans ki ilto') kp1bk; apply.
have bqi : ccw b q (iter i f b) by apply/Knuth_1/red_path; rewrite ?bin.
suff main'' :
 (forall k' : nat, (k' < i)%N -> ccw (iter k' f b) (iter k'.+1 f b) r) ->
  exists t, t \in ts' /\ inside_triangle_seq t r.
  have bnr : b != r by rewrite -/(iter 0 f b) eq_sym rniter.
  have /Knuth_3[ibr | bir] : uniq [:: iter i f b; b; r].
    rewrite /= !inE negb_or bnr eq_sym -{1}/(iter 0 f b) iter_b_neq //.
    by rewrite eq_sym rniter.
    have lower : inside_triangle_seq [:: b; q; iter i f b] r.
      rewrite /=.
      have bqr : ccw b q r by rewrite -nfb; apply (xh b).
      have qir : ccw q (iter i f b) r.
        by rewrite -nfq; apply: xh; rewrite orbit0.
      by rewrite bqr qir ibr.
    case: (main i (leqnn _) lower); first by [].
    by [].   
  by apply: main''; have := convex_hull_split ilto' bir rniter blues_r.
move=> main'.
have rnotins : r \notin inp.
  by apply/(contraNN _ rnotin); rewrite !inE orbC=> ->.
have : inh f p r.
  move=> y; rewrite -(same_orbit bin)=> /trajectP.
  rewrite -size_orbit size_orbit_stable // => [][k] klto ->.
  rewrite -iterS; have [ilek | ] := boolP (i <= k).
    by apply: blues_r; rewrite ilek.
  by rewrite -ltnNge; apply: main'.
move=> /cov [rinp |]; first by move: rnotin; rewrite inE rinp orbT.
case=> t [tin rint]; exists t; split=> //.
by rewrite /ts' add_red_triangles_exact mem_cat tin orbT.
Qed.

Lemma uniq_tr_outside_hull :
  uniq (add_red_triangles ccf q (map_orbit eq_op m b b (map_size m) [:: b]) tr).
Proof.
move: (cot)=> [] trP [] ctr [] incP [] hulP [] no_i [] cov [] unitr cyc.
rewrite add_red_triangles_exact cat_uniq unitr andbT andbC; apply/andP; split.
  apply/hasPn=> t ttr; apply/negP=> /mapP[] j _ t_dec.
  move: uni; rewrite cons_uniq=>/andP[]/negP[]; rewrite -ctr.
  by rewrite (perm_flatten_mem (perm_to_rem ttr)) /= mem_cat t_dec !subE.
rewrite map_inj_in_uniq; first by rewrite rev_uniq iota_uniq.
move=> j1 j2; rewrite !mem_rev !mem_iota add0n !leq0n !andTb=> j1i j2i [] _.
have/sub_orbit_uniq: 0 + i <= size (orbit f p) by rewrite add0n ltnW // ilto'.
move=> /(_ 0)/(uniqP p); rewrite size_map size_iota=> /(_ _ _ j1i j2i).
by rewrite !(nth_map 0) ?size_iota // !nth_iota // !add0n !addn0.
Qed.

Lemma cycle_outside_hull : fcycle nf (orbit nf q).
Proof. by rewrite -/nf orbit_nf cycle_orbit_nf_val. Qed.

Lemma correct_outside_hull :
  correct_triangulation (q :: inp) 
  (add_red_triangles ccf q (map_orbit eq_op m b b (map_size m) [:: b]) tr)
  (new_map m b (iter i f b) q) q.
Proof.
split; first by apply: all_triangles_outtside_hull.
split; first by apply: ctr_outside_hull.
split; first by apply: orbit_nf_inc.
split; first by apply: new_hulP.
split; first by apply: no_i_outside_hull.
split; first by apply: cover_hull_outside_hull.
split; first by apply: uniq_tr_outside_hull.
by apply: cycle_outside_hull.
Qed.

End fixed_red_subpath.

End outside_hull.

Lemma main_correct inp : uniq inp -> 2 < size inp ->
   correct_triangulation inp (naive5 inp).1.1 (naive5 inp).1.2 (naive5 inp).2.
Proof.
elim: inp => [// | q inp IH] uni; rewrite leq_eqVlt=> /orP[/eqP sq3|].
  have:= base_case (esym sq3) uni; case: (naive5 _) => [[tr hf] bp].
  by move=>/(_ _ _ _ erefl).
rewrite /= ltnS=> s3; have -> : (size inp < 3) = false.
  by move: s3; rewrite leqNgt=> /negbTE.
have uni': uniq inp by move: uni; rewrite cons_uniq=> /andP[].
have qnotin : q \notin inp by move: uni; rewrite cons_uniq=> /andP[].
move: (IH uni' s3); case: (naive5 inp)=> [[tr m] p] /= cot.
case h : (pick_triangle_seq tr q) => /= [[t tr'] |].
  have /andP[ins]:= pick_triangle_seq_someP h; rewrite perm_sym=> pr.
  by apply: (correct_inside_triangle s3 uni cot ins pr).
have no_triangle := pick_triangle_seq_noneP h.
have [red_point redo redP]:=
   no_triangle_red_edge no_triangle s3 uni' qnotin cot.
have [b [i [bin [iint [reds blues]]]]] :=
    partition_red_blue s3 uni cot redo redP.
have -> := find_purple_points_mapP s3 uni cot bin iint reds blues.
by apply: (correct_outside_hull s3 uni cot bin iint reds blues).
Qed.

End proofs_in_finType.